import abc
import unittest
import numpy as np
import six

from kafe2.core.minimizers import AVAILABLE_MINIMIZERS

from kafe2.config import kc

from kafe2.fit import IndexedFit
from kafe2.fit.indexed.cost import IndexedCostFunction_Chi2

from kafe2.test.fit.test_fit import AbstractTestFit


def simple_chi2(data, model):
    return np.sum((data - model)**2)

def simple_chi2_explicit_model_name(data, simple_indexed_model):
    return np.sum((data - simple_indexed_model)**2)

def simple_indexed_model(a=1.1, b=2.2, c=3.3):
    x = np.arange(10)
    return a * x ** 2 + b * x + c

def line_indexed_model(a=3.0, b=0.0):
    x = np.arange(10)
    return a * x + b


def analytic_solution(des_mat, cov_mat_inv, data):
    return (
        np.squeeze(np.asarray(np.linalg.inv(
            des_mat.T.dot(cov_mat_inv).dot(des_mat)
        ).dot(des_mat.T)
         .dot(cov_mat_inv)
         .dot(data))))


class TestIndexedFitBasicInterface(AbstractTestFit, unittest.TestCase):

    MINIMIZER = 'scipy'

    def setUp(self):
        self._n_points = 10

        # "jitter" for data smearing
        self._jitter = np.array([-0.3193475 , -1.2404198 , -1.4906926 , -0.78832446,
                                   -1.7638106,   0.36664261,  0.49433821,  0.0719646,
                                    1.95670326,  0.31200215])
        assert len(self._jitter) == self._n_points

        # reference initial values
        self._ref_initial_pars = np.array([1.1, 2.2, 3.3])
        self._ref_initial_model = simple_indexed_model(*self._ref_initial_pars)

        # fit data
        self._ref_data = self._ref_initial_model + self._jitter

        self._linear_design_matrix = np.array([
            np.arange(10) ** 2,
            np.arange(10),
            np.ones(10)
        ]).T

        # pre-fit cost value
        self._ref_initial_cost = simple_chi2(
            data=self._ref_data,
            model=self._ref_initial_model,
        )

        # reference matrices/errors
        self._ref_error = np.ones(self._n_points)
        self._ref_matrix_eye = np.eye(self._n_points)

        # reference fit result values
        #self._nominal_fit_result_pars = (1.1351433, 2.13736919, 2.33346549)
        self._nominal_fit_result_pars = analytic_solution(
            self._linear_design_matrix,
            np.linalg.inv(self._ref_matrix_eye),
            self._ref_data,
        )

        self._nominal_fit_result_model = simple_indexed_model(*self._nominal_fit_result_pars)
        self._nominal_fit_result_cost = simple_chi2(
            data=self._ref_data,
            model=self._nominal_fit_result_model,
        )

        # helper dict with all reference property values
        self._ref_prop_dict = dict(
            did_fit=False,
            model_count=1,

            parameter_values=self._ref_initial_pars,
            parameter_names=('a', 'b', 'c'),
            cost_function_value=self._ref_initial_cost,

            data=self._ref_data,
            data_error=self._ref_error,
            data_cov_mat=self._ref_matrix_eye,
            data_cov_mat_inverse=self._ref_matrix_eye,
            data_cor_mat=self._ref_matrix_eye,

            model=self._ref_initial_model,
            model_error=self._ref_error * 0,
            model_cov_mat=self._ref_matrix_eye * 0,
            model_cov_mat_inverse=None,
            model_cor_mat=self._ref_matrix_eye * np.nan,

            total_error=self._ref_error,
            total_cov_mat=self._ref_matrix_eye,
            total_cov_mat_inverse=self._ref_matrix_eye,
            total_cor_mat=self._ref_matrix_eye,
        )

    def _get_fit(
            self, model_function=None, cost_function=None, errors=None,
            dynamic_error_algorithm=None):
        '''convenience'''
        model_function = model_function or simple_indexed_model
        # TODO: fix default
        cost_function = cost_function or IndexedCostFunction_Chi2(
            errors_to_use='covariance')
        errors = errors or [dict(err_val=1.0)]
        dynamic_error_algorithm = dynamic_error_algorithm or "nonlinear"

        _fit = IndexedFit(
            data=self._ref_data,
            model_function=model_function,
            cost_function=cost_function,
            minimizer=self.MINIMIZER,
            dynamic_error_algorithm=dynamic_error_algorithm
        )
        for _err in errors:
            _fit.add_error(**_err)

        return _fit

    def _get_test_fits(self):
        return {
            'default': self._get_fit(),
            'explicit': self._get_fit(cost_function=simple_chi2),
            'explicit_model': self._get_fit(cost_function=simple_chi2_explicit_model_name),
            'relative_errors_data': self._get_fit(
                errors=[dict(err_val=1.0 / self._ref_data, relative=True, reference='data')]
            ),
        }

    def test_initial_state(self):
        self.run_test_for_all_fits(
            self._ref_prop_dict
        )

    def test_fit_results(self):
        self.run_test_for_all_fits(
            dict(self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars,
                model=self._nominal_fit_result_model,
                did_fit=True,
                cost_function_value=self._nominal_fit_result_cost,
            ),
            call_before_fit=lambda f: f.do_fit(),
            rtol=1e-2
        )

    def test_set_all_parameter_values(self):
        self.run_test_for_all_fits(
            dict(self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars,
                data=self._ref_data,
                model=self._nominal_fit_result_model,
                parameter_names=('a', 'b', 'c'),
                did_fit=False,
                model_count=1,
                cost_function_value=self._nominal_fit_result_cost
            ),
            # set parameters to their final values by hand
            call_before_fit=lambda f: f.set_all_parameter_values(
                self._nominal_fit_result_pars),
        )

    def test_set_all_parameter_values_wrong_number_raise(self):
        with self.assertRaises(ValueError):
            self._get_fit().set_all_parameter_values((1,))
        with self.assertRaises(ValueError):
            self._get_fit().set_all_parameter_values((1,2,3,4,5))

    def test_parameter_defaults(self):
        def dummy_model(a, b, c):
            return np.arange(10)

        _fit = self._get_fit(model_function=dummy_model)

        self._assert_fit_properties(
            _fit,
            dict(
                parameter_values=np.array([1, 1, 1]),
            )
        )

    def test_parameter_partial_defaults(self):
        def dummy_model(a, b, c=3.3):
            return np.arange(10)

        _fit = self._get_fit(model_function=dummy_model)

        self._assert_fit_properties(
            _fit,
            dict(
                parameter_values=np.array([1, 1, 3.3]),
            )
        )

    def test_update_data(self):
        _fit = self._get_fit()

        _fit.data = self._ref_data * 2
        _fit.add_error(err_val=1.0)

        self._assert_fit_properties(
            _fit,
            dict(
                data=self._ref_data * 2,
            )
        )

        _fit.do_fit()
        _new_estimates = np.array(self._nominal_fit_result_pars) * 2

        self._assert_fit_properties(
            _fit,
            dict(
                data=self._ref_data * 2,
                parameter_values=_new_estimates,
            ),
            rtol=1e-2
        )

    def test_update_data_length_mismatch_raise(self):
        # TODO: update when different length limitation is fixed
        _fit = self._get_fit()
        with self.assertRaises(ValueError):
            _fit.data = np.arange(self._n_points + 1)

    def test_reserved_parameter_names_raise(self):
        def dummy_model(data):
            pass

        with self.assertRaises(ValueError) as _exc:
            IndexedFit(data=self._ref_data,
                  model_function=dummy_model,
                  minimizer=self.MINIMIZER)

        self.assertIn('reserved', _exc.exception.args[0])
        self.assertIn('data', _exc.exception.args[0])

    def dummy_model_no_pars_raise(self):
        def dummy_model():
            pass

        with self.assertRaises(ValueError) as _exc:
            IndexedFit(data=self._ref_data,
                  model_function=dummy_model,
                  minimizer=self.MINIMIZER)

        self.assertIn(
            'needs at least one parameter',
            _exc.exception.args[0])

    def dummy_model_varargs_raise(self):
        # TODO: raise even without 'par'
        def dummy_model(x, par, *varargs):
            pass

        with self.assertRaises(ValueError) as _exc:
            IndexedFit(data=self._ref_data,
                  model_function=dummy_model,
                  minimizer=self.MINIMIZER)

        self.assertIn('variable', _exc.exception.args[0])
        self.assertIn('varargs', _exc.exception.args[0])

    def dummy_model_varkwargs_raise(self):
        # TODO: raise even without 'par'
        def dummy_model(x, par, **varkwargs):
            pass

        with self.assertRaises(ValueError) as _exc:
            IndexedFit(data=self._ref_data,
                  model_function=dummy_model,
                  minimizer=self.MINIMIZER)

        self.assertIn('variable', _exc.exception.args[0])
        self.assertIn('varkwargs', _exc.exception.args[0])

    def dummy_model_varargs_varkwargs_raise(self):
        # TODO: raise even without 'par'
        def dummy_model(x, par, *varargs, **varkwargs):
            pass

        with self.assertRaises(ValueError) as _exc:
            IndexedFit(data=self._ref_data,
                  model_function=dummy_model,
                  minimizer=self.MINIMIZER)

        self.assertIn('variable', _exc.exception.args[0])
        self.assertIn('varargs', _exc.exception.args[0])
        # TODO: enable when implemented
        #self.assertIn('varkwargs', _exc.exception.args[0])

    def test_report_before_fit(self):
        # TODO: check report content
        _buffer = six.StringIO()
        _fit = self._get_fit()
        _fit.report(output_stream=_buffer)
        self.assertNotEqual(_buffer.getvalue(), "")

    def test_report_after_fit(self):
        # TODO: check report content
        _buffer = six.StringIO()
        _fit = self._get_fit()
        _fit.do_fit()
        _fit.report(output_stream=_buffer)
        self.assertNotEqual(_buffer.getvalue(), "")

    def test_relative_model_error_nonlinear(self):
        _fit_data_err = self._get_fit(errors=[
            dict(err_val=1.0, relative=False, reference="data"),
            dict(err_val=0.1, relative=True, reference="data")
        ])
        _fit_model_err = self._get_fit(errors=[
            dict(err_val=1.0, relative=False, reference="data"),
            dict(err_val=0.1, relative=True, reference="model")
        ], dynamic_error_algorithm="nonlinear")
        _fit_data_err.do_fit()
        _fit_model_err.do_fit()
        self._assert_fit_results_equal(_fit_data_err, _fit_model_err, rtol=1e-2, atol=2e-2)

    def test_relative_model_error_iterative(self):
        _fit_data_err = self._get_fit(errors=[
            dict(err_val=1.0, relative=False, reference="data"),
            dict(err_val=0.1, relative=True, reference="data")
        ])
        _fit_model_err = self._get_fit(errors=[
            dict(err_val=1.0, relative=False, reference="data"),
            dict(err_val=0.1, relative=True, reference="model")
        ], dynamic_error_algorithm="iterative")
        _fit_data_err.do_fit()
        _fit_model_err.do_fit()
        self._assert_fit_results_equal(_fit_data_err, _fit_model_err, rtol=1e-2)

    def test_fit_scalar_model_function(self):
        def _model_function(a):
            return a

        _fit = IndexedFit([1, 2], _model_function)
        _fit.add_error(1.0)
        _fit.do_fit()
        self.assertAlmostEqual(_fit.parameter_values[0], 1.5)
        _fit.report()


class TestIndexedFitWithSimpleErrors(AbstractTestFit, unittest.TestCase):

    MINIMIZER = 'scipy'

    def setUp(self):
        six.get_unbound_function(TestIndexedFitBasicInterface.setUp)(self)

    def _get_fit(self, errors=None):
        '''convenience'''

        errors = errors or [dict(err_val=1.0)]

        _fit = IndexedFit(
            data=self._ref_data,
            model_function=simple_indexed_model,
            cost_function=IndexedCostFunction_Chi2(errors_to_use='covariance'),
            minimizer=self.MINIMIZER
        )

        for _err in errors:
            _fit.add_error(**_err)

        return _fit

    def _get_test_fits(self):
        return {
            'default': self._get_fit(),
            'two_errors': self._get_fit(
                errors=[
                    dict(err_val=1.0/np.sqrt(2)),
                    dict(err_val=1.0/np.sqrt(2))
                ]
            ),
            'named_errors': self._get_fit(
                errors=[
                    dict(err_val=1.0/np.sqrt(2), name="MyDataError",
                         correlation=0, relative=False, reference='data'),
                    dict(err_val=1.0/np.sqrt(2), name="MyModelError",
                         correlation=0, relative=False, reference='model'),
                ]
            )
        }

    def test_initial_state(self):
        self.run_test_for_all_fits(
            self._ref_prop_dict,
            fit_names=['default', 'two_errors'],
        )

    def test_fit_results(self):
        self.run_test_for_all_fits(
            dict(self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars,
                model=self._nominal_fit_result_model,
                did_fit=True,
                cost_function_value=self._nominal_fit_result_cost,
            ),
            fit_names=['default', 'two_errors'],
            call_before_fit=lambda f: f.do_fit(),
            rtol=1e-2
        )

    def test_get_matching_error_all(self):
        _fit = self._get_test_fits()['named_errors']
        for _mc in (None, dict()):
            _errs = _fit.get_matching_errors(matching_criteria=_mc)
            self.assertEqual(len(_errs), 2)
            self.assertIs(_fit.data_container._error_dicts['MyDataError']['err'], _errs['MyDataError'])
            self.assertIs(_fit._param_model._error_dicts['MyModelError']['err'], _errs['MyModelError'])

    def test_get_matching_error_name(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(name='MyDataError'))
        self.assertEqual(len(_errs), 1)
        self.assertIs(_fit.data_container._error_dicts['MyDataError']['err'], _errs['MyDataError'])

    def test_get_matching_error_type_simple(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(type='simple'))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts['MyDataError']['err'], _errs['MyDataError'])
        self.assertIs(_fit._param_model._error_dicts['MyModelError']['err'], _errs['MyModelError'])

    def test_get_matching_error_type_matrix(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(type='matrix'))
        self.assertEqual(len(_errs), 0)

    def test_get_matching_error_uncorrelated(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(correlated=True))
        self.assertEqual(len(_errs), 0)

    def test_get_matching_error_correlated(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(correlated=False))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts['MyDataError']['err'], _errs['MyDataError'])
        self.assertIs(_fit._param_model._error_dicts['MyModelError']['err'], _errs['MyModelError'])

    def test_get_matching_error_reference_data(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(reference='data'))
        self.assertEqual(len(_errs), 1)
        self.assertIs(_fit.data_container._error_dicts['MyDataError']['err'], _errs['MyDataError'])

    def test_get_matching_error_reference_model(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(reference='model'))
        self.assertEqual(len(_errs), 1)
        self.assertIs(_fit._param_model._error_dicts['MyModelError']['err'], _errs['MyModelError'])


class TestIndexedFitWithMatrixErrors(AbstractTestFit, unittest.TestCase):

    MINIMIZER = 'scipy'

    def setUp(self):
        six.get_unbound_function(TestIndexedFitBasicInterface.setUp)(self)

    def _get_fit(self, errors=None):
        '''convenience'''

        _fit = IndexedFit(
            data=self._ref_data,
            model_function=simple_indexed_model,
            cost_function=IndexedCostFunction_Chi2(
                errors_to_use='covariance'),
            minimizer=self.MINIMIZER
        )

        if errors is None:
            _fit.add_matrix_error(err_matrix=np.eye(self._n_points), matrix_type='cov')
        else:
            for _err in errors:
                if 'err_matrix' in _err:
                    _fit.add_matrix_error(**_err)
                else:
                    _fit.add_error(**_err)

        return _fit

    def _get_test_fits(self):
        return {
            'default': self._get_fit(),
            'cor_matrix_and_error_vector': self._get_fit(
                errors=[
                    dict(err_matrix=np.eye(self._n_points),
                         matrix_type='cor',
                         err_val=1.0
                    )
                ]
            ),
            'two_matrix_errors': self._get_fit(
                errors=[
                    dict(err_matrix=np.eye(self._n_points)/2, matrix_type='cov'),
                    dict(err_matrix=np.eye(self._n_points)/2, matrix_type='cov')
                ]
            ),
            'one_matrix_one_simple_error': self._get_fit(
                errors=[
                    dict(err_matrix=np.eye(self._n_points)/2, matrix_type='cov'),
                    dict(err_val=1/np.sqrt(2))
                ]
            ),
            'named_errors': self._get_fit(
                errors=[
                    dict(err_val=1.0/np.sqrt(2), name="MySimpleDataError",
                         correlation=0, relative=False, reference='data'),
                    dict(err_val=1.0/np.sqrt(2), name="MySimpleModelError",
                         correlation=0, relative=False, reference='model'),
                    dict(err_matrix=np.eye(self._n_points)/2, matrix_type='cov',
                         name="MyMatrixDataError",
                         relative=False, reference='data'),
                    dict(err_matrix=np.eye(self._n_points)/2, matrix_type='cov',
                         name="MyMatrixModelError",
                         relative=False, reference='model'),
                ]
            )
        }

    def test_initial_state(self):
        self.run_test_for_all_fits(
            self._ref_prop_dict,
            fit_names=['default', 'cor_matrix_and_error_vector',
                       'one_matrix_one_simple_error', 'two_matrix_errors'],
        )

    def test_fit_results(self):
        self.run_test_for_all_fits(
            dict(self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars,
                model=self._nominal_fit_result_model,
                did_fit=True,
                cost_function_value=self._nominal_fit_result_cost,
            ),
            fit_names=['default', 'cor_matrix_and_error_vector',
                       'one_matrix_one_simple_error', 'two_matrix_errors'],
            call_before_fit=lambda f: f.do_fit(),
            rtol=1e-2
        )

    def test_get_matching_error_all(self):
        _fit = self._get_test_fits()['named_errors']
        for _mc in (None, dict()):
            _errs = _fit.get_matching_errors(matching_criteria=_mc)
            self.assertEqual(len(_errs), 4)
            self.assertIs(_fit.data_container._error_dicts['MySimpleDataError']['err'], _errs['MySimpleDataError'])
            self.assertIs(_fit._param_model._error_dicts['MySimpleModelError']['err'], _errs['MySimpleModelError'])
            self.assertIs(_fit.data_container._error_dicts['MyMatrixDataError']['err'], _errs['MyMatrixDataError'])
            self.assertIs(_fit._param_model._error_dicts['MyMatrixModelError']['err'], _errs['MyMatrixModelError'])

    def test_get_matching_error_name(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(name='MySimpleDataError'))
        self.assertEqual(len(_errs), 1)
        self.assertIs(_fit.data_container._error_dicts['MySimpleDataError']['err'], _errs['MySimpleDataError'])

    def test_get_matching_error_type_simple(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(type='simple'))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts['MySimpleDataError']['err'], _errs['MySimpleDataError'])
        self.assertIs(_fit._param_model._error_dicts['MySimpleModelError']['err'], _errs['MySimpleModelError'])

    def test_get_matching_error_type_matrix(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(type='matrix'))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts['MyMatrixDataError']['err'], _errs['MyMatrixDataError'])
        self.assertIs(_fit._param_model._error_dicts['MyMatrixModelError']['err'], _errs['MyMatrixModelError'])

    def test_get_matching_error_uncorrelated(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(correlated=True))
        self.assertEqual(len(_errs), 0)

    def test_get_matching_error_correlated(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(correlated=False))
        # NOTE: passing 'correlated' only matches 'matrix' errors, irrespective of 'True'/'False' value passed
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts['MySimpleDataError']['err'], _errs['MySimpleDataError'])
        self.assertIs(_fit._param_model._error_dicts['MySimpleModelError']['err'], _errs['MySimpleModelError'])

    def test_get_matching_error_reference_data(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(reference='data'))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts['MySimpleDataError']['err'], _errs['MySimpleDataError'])
        self.assertIs(_fit.data_container._error_dicts['MyMatrixDataError']['err'], _errs['MyMatrixDataError'])

    def test_get_matching_error_reference_model(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(reference='model'))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit._param_model._error_dicts['MySimpleModelError']['err'], _errs['MySimpleModelError'])
        self.assertIs(_fit._param_model._error_dicts['MyMatrixModelError']['err'], _errs['MyMatrixModelError'])



import yaml
import re
import numpy as np

from ....core.error import SimpleGaussianError, MatrixGaussianError
from ...xy import XYContainer

__all__ = ["add_error_to_container", "write_errors_to_yaml", "process_error_sources",
           "MatrixYamlDumper", "MatrixYamlLoader"]


_yaml_error_section_for_axis = {0: 'x_errors',
                                1: 'y_errors',
                                None: 'errors'}


def add_error_to_container(err_type, container_obj, **kwargs):
    # TODO: check kwargs explicitly
    if err_type == 'simple':
        container_obj.add_error(**kwargs)
    elif err_type == 'matrix':
        container_obj.add_matrix_error(**kwargs)
    else:
        raise TypeError("Unknown error type '{}'. "
                         "Valid: {}".format(err_type, ('simple', 'matrix')))
    return container_obj


def write_errors_to_yaml(container, yaml_doc):
    # TODO: create public error retrieval interface
    for _err_name, _err_dict in container._error_dicts.items():

        _err_obj = _err_dict['err']
        _err_axis = _err_dict.get('axis', None)

        # get the relevant error section for the current axis, creating it if necessary
        _yaml_section = yaml_doc.setdefault(_yaml_error_section_for_axis[_err_axis], [])

        # -- check for relative errors
        _is_relative = _err_obj.relative
        if _is_relative:
            _err_val = _err_obj.error_rel
        else:
            _err_val = _err_obj.error

        # -- collapse identical errors to one float
        if np.allclose(_err_val[0], _err_val):
            _err_val = float(_err_val[0])
        else:
            _err_val = _err_val.tolist()

        # -- handle different error types
        # TODO shouldn't each error be wrapped inside an 'error' namespace?
        if _err_obj.__class__ is SimpleGaussianError:
            _yaml_section.append(dict(name=_err_name,
                                      type='simple',
                                      error_value=_err_val,
                                      relative=_is_relative,
                                      correlation_coefficient=_err_obj._corr_coeff,
                                      # TODO: public interface for _corr_coeff!
                                      )
                                 )
        elif _err_obj.__class__ is MatrixGaussianError:
            _mtype = _err_obj._matrix_type_at_construction  # TODO: public interface!
            _yaml_section.append(dict(name=_err_name,
                                      type='matrix',
                                      matrix_type=_mtype,
                                      relative=_is_relative,
                                      )
                                 )
            if _mtype == 'covariance':
                if _is_relative:
                    _yaml_section[-1]['matrix'] = _err_obj.cov_mat_rel  # .tolist()
                else:
                    _yaml_section[-1]['matrix'] = _err_obj.cov_mat  # .tolist()
            elif _mtype == 'correlation':
                _yaml_section[-1]['matrix'] = _err_obj.cor_mat  # .tolist()
                _yaml_section[-1]['error_value'] = _err_val
            else:
                raise TypeError("Unknown error matrix type '{}'. "
                                "Valid: 'correlation' or 'covariance'.")
        else:
            raise TypeError("No representation for error type {} "
                             "implemented!".format(type(_err_obj)))

    return yaml_doc


def process_error_sources(container_obj, yaml_doc):
    # -- process error sources
    # errors can be specified as a single float, a list of floats, or a kafe2 error object
    # lists of the above are also valid, if the error object is not a list
    if isinstance(container_obj, XYContainer):  # also applies for XYParamModel
        _xerrs = yaml_doc.pop('x_errors', [])
        if isinstance(_xerrs, (int, float, str)):
            _xerrs = [_xerrs] * container_obj.size
        if len(_xerrs) > 0 and isinstance(_xerrs[0], (int, float, str)):
            _xerrs = [_xerrs]
        _yerrs = yaml_doc.pop('y_errors', [])
        if not isinstance(_yerrs, list):
            _yerrs = [_yerrs] * container_obj.size
        if len(_yerrs) > 0 and isinstance(_yerrs[0], (int, float, str)):
            _yerrs = [_yerrs]
        _errs = _xerrs + _yerrs
        _axes = [0] * len(_xerrs) + [1] * len(_yerrs)  # 0 for 'x', 1 for 'y'
    else:
        _errs = yaml_doc.pop('errors', [])
        if not isinstance(_errs, list):
            _errs = [_errs] * container_obj.size
        if len(_errs) > 0 and isinstance(_errs[0], float):
            _errs = [_errs]
        _axes = [None] * len(_errs)

    # add error sources, if any
    for _err, _axis in zip(_errs, _axes):
        # If error is a list, check for mixture of relative/absolute errors.
        # Add it as a simple error and don't try to interpret it as a kafe2 error object
        if isinstance(_err, list):
            _rel = np.zeros(len(_err))
            _abs = np.zeros(len(_err))
            for i, _val in enumerate(_err):
                if isinstance(_val, str) and _val.endswith("%"):
                    try:
                        _rel_err_percent = float(_val[:-1])
                    except ValueError:
                        raise ValueError("Cannot convert string to relative error: %s" % _err)
                    _rel[i] = _rel_err_percent
                else:
                    try:
                        _abs_err = float(_val)
                    except ValueError:
                        raise ValueError("Cannot convert {} to error value.".format(_val))
                    _abs[i] = _val
            if _axis is not None:
                container_obj = add_error_to_container('simple', container_obj,
                                                       err_val=0.01 * _rel,
                                                       relative=True,
                                                       axis=_axis)
                container_obj = add_error_to_container('simple', container_obj,
                                                       err_val=_abs,
                                                       relative=False,
                                                       axis=_axis)
            else:
                container_obj = add_error_to_container('simple', container_obj,
                                                       err_val=0.01 * _rel,
                                                       relative=True)
                container_obj = add_error_to_container('simple', container_obj,
                                                       err_val=_abs,
                                                       relative=False)
            continue
        elif isinstance(_err, (float, int, str)):
            raise ValueError("Failed to read in errors: {}".format(_err))

        _add_kwargs = dict()
        # translate and check that all required keys are present
        try:
            _err_type = _err.get('type', 'simple')

            _add_kwargs['name'] = _err.get('name', None)

            if _err_type == 'simple':
                _add_kwargs['err_val'] = _err['error_value']
                _add_kwargs['correlation'] = _err.get('correlation_coefficient', 0)
            elif _err_type == 'matrix':
                _add_kwargs['err_matrix'] = _err['matrix']
                _add_kwargs['matrix_type'] = _err['matrix_type']
                # default None only mandatory for cor mats; check done later
                _add_kwargs['err_val'] = _err.get('error_value', None)
            else:
                raise TypeError("Unknown error type '{}'. "
                                 "Valid: {}".format(_err_type, ('simple', 'matrix')))

            _add_kwargs['relative'] = _err.get('relative', False)

            # if needed, specify the axis (only for 'xy' containers)
            if _axis is not None:
                _add_kwargs['axis'] = _axis
        except KeyError as e:
            # KeyErrors mean the YAML is incomplete -> raise
            raise ValueError("Missing required key '%s' for error specification" % e.args[0])

        # add error to data container
        container_obj = add_error_to_container(_err_type, container_obj, **_add_kwargs)

    return container_obj, yaml_doc


class MatrixYamlDumper(yaml.Dumper):
    """Custom directives for writing out matrices"""

    _regex_space_after_newline = re.compile(r"\n\s+")  # match all spaces immediately after newline
    # match exactly 'n' sequences of non-whitespace followed by optional whitespace
    _regex_format_n_numbers = r"((\S+\s*){{}})"

    def matrix(self, numpy_matrix):
        """Represent a matrix as a space/endline separated string.
        Note: symmetric matrices are represented as a lower triangular matrix by default
        """
        _is_symmetric = np.allclose(numpy_matrix, numpy_matrix.T)

        # remove brackets
        _string_repr = str(numpy_matrix).replace('[[', ' [').replace(']]', '] ').replace('[', '').\
            replace(']', '').strip()
        # remove all spaces immediately after newline
        _string_repr = re.sub(self.__class__._regex_space_after_newline, "\n", _string_repr)

        # if symmetric, remove everything above upper diagonal
        if _is_symmetric and False:
            _rows = _string_repr.split('\n')
            for _irow, _row in enumerate(_rows):
                _rows[_irow] = re.sub(r"^((\S+\s*){{{}}}).*$".format(_irow+1), r"\1", _row).strip()
            _string_repr = "\n".join(_rows)
            # write lower triangular matrix using the '|'-style
            return self.represent_scalar('!symmetric_matrix', _string_repr, style='|')

        # write full matrix using the '|'-style
        return self.represent_scalar('!matrix', _string_repr, style='|')


# representers for covariance matrices errors
MatrixYamlDumper.add_representer(np.array, MatrixYamlDumper.matrix)


class MatrixYamlLoader(yaml.Loader):
    """custom directives for reading in matrices"""

    def matrix(self, node):
        """construct a matrix from a tab/endline separated string"""
        _rows = node.value.split('\n')
        _mat = [_row.split() for _row in _rows if _row]  # last row may be empty -> leave out
        return np.array(_mat, dtype=float)

    def symmetric_matrix(self, node):
        """construct a lower triangular matrix from a tab/endline separated string"""
        _rows = node.value.split('\n')[:-1]  # last endline ends the matrix -> no new row
        _mat = [_row.split() for _row in _rows]
        _np_mat = np.array(np.zeros((len(_rows), len(_rows))), dtype=float)

        for _i, _row in enumerate(_mat):
            # check shape -> row index must match row length
            if len(_row) != _i + 1:
                raise ValueError("Cannot parse lower triangular matrix: "
                                 "row #{} should have length {}, got {} "
                                 "instead!".format(_i+1, _i+1, len(_row)))
            # fill matrix
            _np_mat[_i, 0:len(_row)] = _row   # fill below diagonal

        _np_mat = _np_mat + np.tril(_np_mat, -1).T  # symmetrize

        return _np_mat


MatrixYamlLoader.add_constructor('!matrix', MatrixYamlLoader.matrix)
MatrixYamlLoader.add_constructor('!symmetric_matrix', MatrixYamlLoader.symmetric_matrix)



try:
    import typing  # help IDEs with type-hinting inside docstrings
except ImportError:
    pass
import numpy  # help IDEs with type-hinting inside docstrings
from collections import OrderedDict
from copy import deepcopy
import numpy as np

from ...core.error import CovMat
from ...tools import print_dict_as_table
from .._base import FitBase, DataContainerBase, ModelFunctionBase
from .container import XYContainer
from .cost import XYCostFunction_Chi2, STRING_TO_COST_FUNCTION
from .model import XYParametricModel
from .plot import XYPlotAdapter
from ..util import function_library, add_in_quadrature, invert_matrix


__all__ = ['XYFit']


class XYFit(FitBase):
    CONTAINER_TYPE = XYContainer
    MODEL_TYPE = XYParametricModel
    MODEL_FUNCTION_TYPE = ModelFunctionBase
    PLOT_ADAPTER_TYPE = XYPlotAdapter
    RESERVED_NODE_NAMES = {'y_data', 'y_model', 'cost',
                           'x_error', 'y_data_error', 'y_model_error', 'total_error',
                           'x_cov_mat', 'y_data_cov_mat', 'y_model_cov_mat', 'total_cov_mat',
                           'x_cor_mat', 'y_data_cor_mat', 'y_model_cor_mat', 'total_cor_mat',
                           'x_cov_mat_inverse', 'y_data_cov_mat_inverse', 'y_model_cov_mat_inverse', 'total_cor_mat_inverse'
                           'x_data_cov_mat'}
    _BASIC_ERROR_NAMES = {
        'x_data_error', 'x_model_error', 'x_data_cov_mat', 'x_model_cov_mat',
        'y_data_error', 'y_model_error', 'y_data_cov_mat', 'y_model_cov_mat'
    }
    X_ERROR_ALGORITHMS = ('iterative linear', 'nonlinear')
    _STRING_TO_COST_FUNCTION = STRING_TO_COST_FUNCTION
    _AXES = (None, "x", "y")
    _MODEL_NAME = "y_model"
    _MODEL_ERROR_NODE_NAMES = ["y_model_error", "y_model_cov_mat"]
    _PROJECTED_NODE_NAMES = ["total_error", "total_cov_mat"]

    def __init__(self,
                 xy_data,
                 model_function=function_library.linear_model,
                 cost_function="chi2",
                 minimizer=None, minimizer_kwargs=None,
                 dynamic_error_algorithm="nonlinear"):
        """Construct a fit of a model to *xy* data.

        :param xy_data: A :py:obj:`~.XYContainer` or a raw 2D array of shape ``(2, N)``
         containing the measurement data.
        :type xy_data: XYContainer or typing.Sequence
        :param model_function: The model function as a native Python function where the first
            argument denotes the independent *x* variable or an already defined
            :py:class:`~kafe2.fit.xy.XYModelFunction` object.
        :type model_function: typing.Callable
        :param cost_function: The cost function this fit uses to find the best parameters.
        :type cost_function: str or typing.Callable
        :param minimizer: The minimizer to use for fitting. Either :py:obj:`None`, ``"iminuit"``,
            ``"tminuit"``, or ``"scipy"``.
        :type minimizer: str or None
        :param minimizer_kwargs: Dictionary with kwargs for the minimizer.
        :type minimizer_kwargs: dict
        """
        super(XYFit, self).__init__(
            data=xy_data, model_function=model_function, cost_function=cost_function,
            minimizer=minimizer, minimizer_kwargs=minimizer_kwargs,
            dynamic_error_algorithm=dynamic_error_algorithm)

    # -- private methods

    def _init_nexus(self):
        super(XYFit, self)._init_nexus()

        self._nexus.add_function(
            func=self._project_cov_mat,
            func_name="total_cov_mat",
            par_names=[
                "x_total_cov_mat",
                "y_total_cov_mat",
                "x_model",
                "parameter_values"
            ],
            existing_behavior="replace"
        )
        self._nexus.add_function(
            func=self._project_error,
            func_name="total_error",
            par_names=[
                "x_total_error",
                "y_total_error",
                "x_model",
                "parameter_values"
            ],
            existing_behavior="replace"
        )

        self._nexus.add_dependency(
            'y_model',
            depends_on=(
                'x_model',
                'parameter_values'
            )
        )

        self._nexus.add_dependency(
            'x_model',
            depends_on=(
                'x_data',
            )
        )

    def _set_new_data(self, new_data):
        if isinstance(new_data, self.CONTAINER_TYPE):
            self._data_container = deepcopy(new_data)
        elif isinstance(new_data, DataContainerBase):
            raise TypeError("Incompatible container type '%s' (expected '%s')"
                                 % (type(new_data), self.CONTAINER_TYPE))
        else:
            _x_data = new_data[0]
            _y_data = new_data[1]
            self._data_container = XYContainer(_x_data, _y_data, dtype=float)
        self._data_container._on_error_change_callback = self._on_error_change

        # update nexus data nodes
        self._nexus.get('x_data').mark_for_update()
        self._nexus.get('y_data').mark_for_update()

    def _set_new_parametric_model(self):
        self._param_model = XYParametricModel(
            self.x_model,
            self._model_function,
            self.parameter_values
        )

    def _report_data(self, output_stream, indent, indentation_level):
        output_stream.write(indent * indentation_level + '########\n')
        output_stream.write(indent * indentation_level + '# Data #\n')
        output_stream.write(indent * indentation_level + '########\n\n')
        _data_table_dict = OrderedDict()
        _data_table_dict['X Data'] = self.x_data
        if self._data_container.has_x_errors:
            _data_table_dict['X Data Error'] = self.x_data_error
            _data_table_dict['X Data Correlation Matrix'] = self.x_data_cor_mat

        print_dict_as_table(_data_table_dict, output_stream=output_stream, indent_level=indentation_level + 1)
        output_stream.write('\n')

        _data_table_dict = OrderedDict()
        _data_table_dict['Y Data'] = self.y_data
        if self._data_container.has_y_errors:
            _data_table_dict['Y Data Error'] = self.y_data_error
            _data_table_dict['Y Data Correlation Matrix'] = self.y_data_cor_mat

        print_dict_as_table(_data_table_dict, output_stream=output_stream, indent_level=indentation_level + 1)
        output_stream.write('\n')

    def _report_model(self, output_stream, indent, indentation_level):
        # call base method to show header and model function
        super(XYFit, self)._report_model(output_stream, indent, indentation_level)
        _model_table_dict = OrderedDict()
        _model_table_dict['X Model'] = self.x_model
        if self._param_model.has_x_errors:
            _model_table_dict['X Model Error'] = self.x_model_error
            _model_table_dict['X Model Correlation Matrix'] = self.x_model_cor_mat

        print_dict_as_table(_model_table_dict, output_stream=output_stream, indent_level=indentation_level + 1)
        output_stream.write('\n')

        _model_table_dict = OrderedDict()
        _model_table_dict['Y Model'] = self.y_model
        if self._param_model.has_y_errors:
            _model_table_dict['Y Model Error'] = self.y_model_error
            _model_table_dict['Y Model Correlation Matrix'] = self.y_model_cor_mat

        print_dict_as_table(_model_table_dict, output_stream=output_stream, indent_level=indentation_level + 1)
        output_stream.write('\n')
        if self._param_model.get_matching_errors({"relative": True, "axis": 1}):
            output_stream.write(indent * (indentation_level + 1))
            output_stream.write(
                "y model covariance matrix was calculated dynamically relative to y model values.\n"
            )
            output_stream.write("\n")

    def _project_cov_mat(self, x_cov_mat, y_cov_mat, x_model, parameter_values):
        _derivatives = self._param_model.eval_model_function_derivative_by_x(
            x=x_model,
            dx=0.01 * np.sqrt(np.diag(x_cov_mat)),
            model_parameters=parameter_values
        )
        return y_cov_mat + x_cov_mat * np.outer(_derivatives, _derivatives)

    def _project_error(self, x_error, y_error, x_model, parameter_values):
        _derivatives = self._param_model.eval_model_function_derivative_by_x(
            x=x_model,
            dx=0.01 * x_error,
            model_parameters=parameter_values
        )
        return np.sqrt(np.square(y_error) + np.square(x_error * _derivatives))

    def _set_data_as_model_ref(self):
        _errs_and_old_refs = []
        for _err in self._param_model.get_matching_errors({"relative": True, "axis": 1}).values():
            _old_ref = _err.reference
            _err.reference = self._data_container.y
            _errs_and_old_refs.append((_err, _old_ref))
        return _errs_and_old_refs

    def _iterative_fits_needed(self):
        return (bool(self._param_model.get_matching_errors({"relative": True, "axis": 1}))
                or self.has_x_errors) \
               and self._dynamic_error_algorithm == "iterative"

    def _second_fit_needed(self):
        return (bool(self._param_model.get_matching_errors({"relative": True, "axis": 1}))
                or self.has_x_errors) and self._dynamic_error_algorithm == "nonlinear"

    def _get_node_names_to_freeze(self, first_fit):
        if self._dynamic_error_algorithm == "iterative" or (first_fit and self.has_x_errors):
            return self._PROJECTED_NODE_NAMES + super(
                XYFit, self)._get_node_names_to_freeze(first_fit)
        else:
            return super(XYFit, self)._get_node_names_to_freeze(first_fit)

    # -- public properties

    @property
    def has_x_errors(self):
        """:py:obj:`True`` if at least one *x* uncertainty source has been defined.

        :rtype: bool
        """
        return self._data_container.has_x_errors or self._param_model.has_x_errors

    @property
    def has_y_errors(self):
        """:py:obj:`True`` if at least one *y* uncertainty source has been defined

        :rtype: bool
        """
        return self._data_container.has_y_errors or self._param_model.has_y_errors

    @property
    def x_data(self):
        """1D array containing the measurement *x* values.

        :rtype: numpy.ndarray[float]
        """
        return self._data_container.x

    @property
    def x_model(self):
        """1D array containing the model *x* values. The same as :py;obj:`.x_data` for an
            :py:obj:`~.XYFit`.

        :rtype: numpy.ndarray[float]
        """
        return self.x_data

    @property
    def y_data(self):
        """1D array containing the measurement *y* values.

        :rtype: numpy.ndarray[float]
        """
        return self._data_container.y

    @property
    def model(self):
        """2D array of shape ``(2, N)`` containing the *x* and *y* model values

        :rtype: numpy.ndarray
        """
        return self._param_model.data

    @property
    def x_data_error(self):
        """1D array containing the pointwise *x* data uncertainties

        :rtype: numpy.ndarray[float]
        """
        return self._data_container.x_err

    @property
    def y_data_error(self):
        """1D array containing the pointwise *y* data uncertainties

        :rtype: numpy.ndarray[float]
        """
        return self._data_container.y_err

    @property
    def data_error(self):
        """1D array containing the pointwise *xy* uncertainties projected onto the *y* axis.

        :rtype: numpy.ndarray[float]
        """
        return self._project_error(
            self.x_data_error, self.y_data_error, self.x_model, self.parameter_values)

    @property
    def x_data_cov_mat(self):
        """2D array of shape ``(N, N)`` containing the data *x* covariance matrix.

        :rtype: numpy.ndarray
        """
        return self._data_container.x_cov_mat

    @property
    def y_data_cov_mat(self):
        """2D array of shape ``(N, N)`` containing the data *y* covariance matrix.

        :rtype: numpy.ndarray
        """
        return self._data_container.y_cov_mat

    @property
    def data_cov_mat(self):
        """2D array of shape ``(N, N)`` containing the data *xy* covariance matrix (projected
        onto the *y* axis).

        :rtype: numpy.ndarray
        """
        return self._project_cov_mat(
            self.x_data_cov_mat, self.y_data_cov_mat, self.x_model, self.parameter_values)

    @property
    def x_data_cov_mat_inverse(self):
        """2D array of shape ``(N, N)`` containing the inverse of the data *x* covariance matrix or
        :py:obj:`None` if singular.

        :rtype: numpy.ndarray or None
        """
        return self._data_container.x_cov_mat_inverse

    @property
    def y_data_cov_mat_inverse(self):
        """2D array of shape ``(N, N)`` containing the inverse of the data *y* covariance matrix or
        :py:obj:`None` if singular.

        :rtype: numpy.ndarray or None
        """
        return self._data_container.y_cov_mat_inverse

    @property
    def data_cov_mat_inverse(self):
        """2D array of shape ``(N, N)`` containing the inverse of the data *xy* covariance matrix
        projected onto the *y* axis. :py:obj:`None` if singular.

        :rtype: numpy.ndarray or None
        """
        return invert_matrix(self.data_cov_mat)

    @property
    def x_data_cor_mat(self):
        """2D array of shape ``(N, N)`` containing the data *x* correlation matrix.

        :rtype: numpy.ndarray
        """
        return self._data_container.x_cor_mat

    @property
    def y_data_cor_mat(self):
        """2D array of shape ``(N, N)`` containing the data *y* correlation matrix.

        :rtype: numpy.ndarray
        """
        return self._data_container.y_cor_mat

    @property
    def data_cor_mat(self):
        """2D array of shape ``(N, N)`` containing the data *xy* correlation matrix projected
        onto the *y* axis.

        :rtype: numpy.ndarray
        """
        return CovMat(self.data_cov_mat).cor_mat

    @property
    def y_model(self):
        """1D array of *y* model predictions for the data points.

        :rtype: numpy.ndarray[float]
        """
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        self._param_model.x = self.x_model
        return self._param_model.y

    @property
    def x_model_error(self):
        """1D array of pointwise model *x* uncertainties.

        :rtype: numpy.ndarray[float]
        """
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        self._param_model.x = self.x_model
        return self._param_model.x_err

    @property
    def y_model_error(self):
        """1D array of pointwise model *y* uncertainties.

        :rtype: numpy.ndarray[float]
        """
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        self._param_model.x = self.x_model
        return self._param_model.y_err

    @property
    def model_error(self):
        """1D array of pointwise model *xy* uncertainties projected onto the *y* axis.

        :rtype: numpy.ndarray[float]
        """
        return self._project_error(
            self.x_model_error, self.y_model_error, self.x_model, self.parameter_values)

    @property
    def x_model_cov_mat(self):
        """2D array of shape ``(N, N)`` containing the model *x* covariance matrix.

        :rtype: numpy.ndarray
        """
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        self._param_model.x = self.x_model
        return self._param_model.x_cov_mat

    @property
    def y_model_cov_mat(self):
        """2D array of shape ``(N, N)`` containing the model *y* covariance matrix.

        :rtype: numpy.ndarray
        """
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        self._param_model.x = self.x_model
        return self._param_model.y_cov_mat

    @property
    def model_cov_mat(self):
        """2D array of shape ``(N, N)`` containing the model *xy* covariance matrix projected onto
        the *y* axis.

        :rtype: numpy.ndarray
        """
        return self._project_cov_mat(
            self.x_model_cov_mat, self.y_model_cov_mat, self.x_model, self.parameter_values)

    @property
    def x_model_cov_mat_inverse(self):
        """2D array of shape ``(N, N)`` containing the inverse of the model *x* covariance matrix
        or :py:obj:`None` if singular.

        :rtype: numpy.ndarray or None
        """
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        self._param_model.x = self.x_model
        return self._param_model.x_cov_mat_inverse

    @property
    def y_model_cov_mat_inverse(self):
        """2D array of shape ``(N, N)`` containing the inverse of the model *y* covariance matrix
        or :py:obj:`None` if singular.

        :rtype: numpy.ndarray
        """
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        self._param_model.x = self.x_model
        return self._param_model.y_cov_mat_inverse

    @property
    def model_cov_mat_inverse(self):
        """2D array of shape ``(N, N)`` containing the inverse of the model *xy* covariance matrix
        projected onto the *y* axis. :py:obj:`None`` if singular.

        :rtype: numpy.ndarray
        """
        return invert_matrix(self.model_cov_mat)

    @property
    def x_model_cor_mat(self):
        """2D array of shape ``(N, N)`` containing the model *x* correlation matrix.

        :rtype: numpy.ndarray
        """
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        self._param_model.x = self.x_model
        return self._param_model.x_cor_mat

    @property
    def y_model_cor_mat(self):
        """2D array of shape ``(N, N)`` containing the model *y* correlation matrix.

        :rtype: numpy.ndarray
        """
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        self._param_model.x = self.x_model
        return self._param_model.y_cor_mat

    @property
    def model_cor_mat(self):
        """2D array of shape ``(N, N)`` containing the model *xy* correlation matrix projected onto
        the *y* axis.

        :rtype: numpy.ndarray
        """
        return CovMat(self.model_cov_mat).cor_mat

    @property
    def x_total_error(self):
        """1D array of total pointwise *x* uncertainties.

        :rtype: numpy.ndarray[float]
        """
        return self._nexus.get("x_total_error").value

    @property
    def y_total_error(self):
        """1D array of total pointwise *y* uncertainties

        :rtype: numpy.ndarray[float]
        """
        return self._nexus.get("y_total_error").value

    @property
    def total_error(self):
        """1D array of the total pointwise *xy* uncertainties projected onto the *y* axis.

        :rtype: numpy.ndarray[float]
        """
        return self._nexus.get("total_error").value

    @property
    def x_total_cov_mat(self):
        """2D array of shape ``(N, N)`` containing the total *x* covariance matrix.

        :rtype: numpy.ndarray
        """
        return self._nexus.get("x_total_cov_mat").value

    @property
    def y_total_cov_mat(self):
        """2D array of shape ``(N, N)`` containing the total *y* covariance matrix.

        :rtype: numpy.ndarray
        """
        return self._nexus.get("y_total_cov_mat").value

    @property
    def total_cov_mat(self):
        """2D array of shape ``(N, N)`` containing the total *xy* covariance matrix projected onto
        the *y* axis.

        :rtype: numpy.ndarray
        """
        return self._nexus.get("total_cov_mat").value

    @property
    def x_total_cov_mat_inverse(self):
        """2D array of shape ``(N, N)`` containing inverse of the total *x* covariance matrix.
        :py:obj:`None` if singular.

        :rtype: numpy.ndarray
        """
        return invert_matrix(self.x_total_cov_mat)

    @property
    def y_total_cov_mat_inverse(self):
        """2D array of shape ``(N, N)`` containing inverse of the total *y* covariance matrix.
        :py:obj:`None` if singular.

        :rtype: numpy.ndarray
        """
        return invert_matrix(self.y_total_cov_mat)

    @property
    def total_cov_mat_inverse(self):
        """2D array of shape ``(N, N)`` containing theinverse of the total *xy* covariance matrix
        projected onto the *y* axis. :py:obj:`None` if singular.

        :rtype: numpy.ndarray
        """
        return invert_matrix(self.total_cov_mat)

    @property
    def x_total_cor_mat(self):
        """2D array of shape ``(N, N)`` containing the total *x* correlation matrix.

        :rtype: numpy.ndarray
        """
        return CovMat(self.x_total_cov_mat).cor_mat

    @property
    def y_total_cor_mat(self):
        """2D array of shape ``(N, N)`` containing the total *y* correlation matrix.

        :rtype: numpy.ndarray
        """
        return CovMat(self.y_total_cov_mat).cor_mat

    @property
    def x_range(self):
        """Minimum and maximum values of the *x* measurement data.

        :rtype: tuple[float, float]
        """
        return self._data_container.x_range

    @property
    def y_range(self):
        """Minimum and maximum values of the *y* measurement data.

        :rtype: tuple[float, float]
        """
        return self._data_container.y_range

    # -- public methods

    def add_error(self, axis, err_val,
                  name=None, correlation=0, relative=False, reference='data'):
        """Add an uncertainty source for an axis to the data container.

        :param axis: ``'x'``/``0`` or ``'y'``/``1``
        :type axis: str or int
        :param err_val: Pointwise uncertainties or a single uncertainty for all data points.
        :type err_val: float or typing.Sequence[float]
        :param name: Unique name for this uncertainty source. If :py:obj:`None`, the name
            of the error source will be set to a random alphanumeric string.
        :type name: str or None
        :param correlation: Correlation coefficient between any two distinct data points.
        :type correlation: float
        :param relative: If :py:obj:`True`, **err_val** will be interpreted as a *relative*
            uncertainty.
        :type relative: bool
        :param reference: Which reference values to use when calculating absolute errors from
            relative errors. Either ``'data'`` or ``'model'``.
        :type reference: str
        :return: An error id uniquely identifying the created error source.
        :rtype: str
        """
        return super(XYFit, self).add_error(err_val=err_val,
                                            name=name,
                                            correlation=correlation,
                                            relative=relative,
                                            reference=reference,
                                            axis=axis)

    def add_matrix_error(self, axis, err_matrix, matrix_type,
                         name=None, err_val=None, relative=False, reference='data'):
        """Add a matrix uncertainty source for an axis to the data container.

        :param axis: ``'x'``/``0`` or ``'y'``/``1``
        :type axis: str or int
        :param err_matrix: 2D array of shape ``(size, size)`` containing the covariance or
            correlation matrix
        :type err_matrix: numpy.ndarray
        :param matrix_type: One of ``'covariance'``/``'cov'`` or ``'correlation'``/``'cor'``.
        :type matrix_type: str
        :param name: Unique name for this uncertainty source. If :py:obj:`None`, the name
            of the error source will be set to a random alphanumeric string.
        :type name: str or None
        :param err_val: The pointwise uncertainties. This is mandatory if only a correlation
            matrix is given.
        :type err_val: typing.Sequence[float]
        :param relative: If :py:obj:`True`, the covariance matrix and/or **err_val** will be
            interpreted as a *relative* uncertainty.
        :type relative: bool
        :param reference: Which reference values to use when calculating absolute errors from
            relative errors. Either ``'data'`` or ``'model'``.
        :type reference: str
        :return: An error id uniquely identifying the created error source.
        :rtype: str
        """
        return super(XYFit, self).add_matrix_error(err_matrix=err_matrix,
                                                   matrix_type=matrix_type,
                                                   name=name,
                                                   err_val=err_val,
                                                   relative=relative,
                                                   reference=reference,
                                                   axis=axis)

    def eval_model_function(self, x=None, model_parameters=None):
        """Evaluate the model function.

        :param x: 1D array containing the values of *x* at which to evaluate the model function. If
            :py:obj:`None`, the data *x* values :py:obj:`~XYFit.x_data` are used.
        :type x: numpy.ndarray[float]
        :param model_parameters: The model parameter values. If :py:obj:`None`, the current values
            :py:obj:`~.XYFit.parameter_values` are used.
        :type model_parameters: typing.Collection[float]
        :return: Model function values at the given *x*-values.
        :rtype: numpy.ndarray[float]
        """
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        self._param_model.x = self.x_model
        return self._param_model.eval_model_function(x=x, model_parameters=model_parameters)

    def eval_model_function_derivative_by_parameters(self, x=None, model_parameters=None, par_dx=None):
        """Evaluate the derivative of the model function with respect to the model parameters.

        :param x: 1D array containing the *x* values at which to evaluate the model function. If
            :py:obj:`None`, the data *x* values :py:obj:`~XYFit.x_data` are used.
        :type x: numpy.ndarray[float]
        :param model_parameters: 1D array containing the model parameter values. If :py:obj:`None`,
            the current values :py:obj:`~XYFit.parameter_values` are used.
        :type model_parameters: typing.Collection[float]
        :param par_dx: 1D array with length ``pars`` containing the numeric differentiation step
            size for each parameter. If :py:obj:`None` and a fit has been performed, 1% of the parameter uncertainties
            is used.
        :type par_dx: typing.Collection[float]
        :return: 2D array of shape ``(par, N)`` containing the model function derivatives for
            each parameter at the given *x* values.
        :rtype: numpy.ndarray[numpy.ndarray[float]]
        """
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        self._param_model.x = self.x_model
        if par_dx is None and self.did_fit:
            par_dx = 1e-2*self.parameter_errors
        return self._param_model.eval_model_function_derivative_by_parameters(
            x=x, model_parameters=model_parameters, par_dx=par_dx)

    def error_band(self, x=None):
        """Calculate the symmetric model uncertainty at every given point *x*. This is only
        possible after a fit has been performed with the :py:meth:`.do_fit` method.

        :param numpy.ndarray[float] x: 1D array containing the values of *x* at which to calculate
            the model uncertainty.
        :return: 1D array containing the model uncertainties at the given *x* values.
        :rtype: numpy.ndarray[float]
        """
        if not self.did_fit:
            raise RuntimeError('Cannot calculate an error band without first performing a fit.')
        if x is None:
            x = self.x_model
        if self.parameter_cov_mat is None:
            return np.zeros_like(x)

        _f_deriv_by_params = self.eval_model_function_derivative_by_parameters(x=x)
        # here: df/dp[par_idx]|x=x[x_idx] = _f_deriv_by_params[par_idx][x_idx]

        _f_deriv_by_params = _f_deriv_by_params.T
        # here: df/dp[par_idx]|x=x[x_idx] = _f_deriv_by_params[x_idx][par_idx]

        _band_y = np.zeros_like(x)

        # Cut out fixed parameters which have nan as derivative:
        _not_pars_fixed = [_par_name not in self._fitter.fixed_parameters
                           for _par_name in self.parameter_names]
        _cut_parameter_cov_mat = self.parameter_cov_mat[_not_pars_fixed][:, _not_pars_fixed]
        for _x_idx, _x_val in enumerate(x):
            _p_res = _f_deriv_by_params[_x_idx, _not_pars_fixed]
            _band_y[_x_idx] = _p_res.dot(_cut_parameter_cov_mat).dot(_p_res)

        return np.sqrt(_band_y)



import warnings
import numpy as np

from copy import deepcopy


def add_pad_to_range(data_range, pad_coeff=1.1, additional_pad=None, scale='linear'):
    """Add padding to a range, so that plots will use more space to the left and right of the
    data point range.

    :param data_range: The range of the data points
    :type data_range: tuple[float, float]
    :param float pad_coeff: The padding coefficient. The default value ``1.1`` adds 10% padding in
        total, meaning five percent at both sides.
    :param additional_pad: A :py:obj:`tuple` containing additional static padding for the lower and
        upper end.
    :type additional_pad: tuple[float, float] or None
    :param str scale: Either ``'linear'`` or ``'log'``. If ``'log'``. The space used for
        calculating the padding.
    :return tuple[float, float]: The data_range including the additional padding.
    """
    if additional_pad is None:
        additional_pad = (0, 0)
    if scale == 'linear':
        _min, _max = data_range
        _w = _max - _min
        return (0.5 * (_min + _max - _w * pad_coeff) - additional_pad[0],
                0.5 * (_min + _max + _w * pad_coeff) + additional_pad[1])
    if scale == 'log':
        _expmin, _expmax = np.log10(data_range)
        _w = _expmax - _expmin
        return (10 ** (0.5 * (_expmin + _expmax - _w * pad_coeff) - additional_pad[0]),
                10 ** (0.5 * (_expmin + _expmax + _w * pad_coeff) + additional_pad[1]))
    warnings.warn("Unknown scale \"{scale}\" when calculating the additional plot padding, "
                  "returning without additional padding.".format(scale=scale))
    return data_range


# represent errors in a binwise fashion as rectangles
def step_fill_between(
        axes, x, y, yerr=None, xerr=None, draw_central_value=False, continuous=True, **kwargs):
    """
    Draw two step functions and fill the area in-between.

    :param axes: target ``Axes`` object
    :type axes: ``matplotlib`` ``Axes`` object
    :param x: *x* coordinates of the steps
    :type x: list of float
    :param y: *y* coordinates of the steps
    :type y: list of float
    :param yerr: distances between the center and upper step function (or lower and upper, if tuple)
    :type yerr: list of float or list of 2-tuple of float
    :param xerr: distances between the step center and preceding/next step (or both, if tuple)
    :type xerr: list of float or list of 2-tuple of float
    :param draw_central_value: if ``True``, also draw the central step function
    :type draw_central_value: bool
    :param kwargs: keyword arguments accepted by ``matplotlib`` methods ``plot`` and ``fill_between``
    :type kwargs: dict
    :return: plot handle(s)
    """
    # -- give errors common structure: tuple(err_dn, err_up)

    _yerr = yerr
    _xerr = xerr

    if _yerr is not None:
        if isinstance(_yerr, tuple) and len(_yerr)==1:
            _yerr = (yerr[0], yerr[0])
        elif isinstance(_yerr, tuple) and len(_yerr)==2:
            _yerr = (yerr[0], yerr[1])
        else:
            try:
                if _yerr.ndim == 1:
                    _yerr = (yerr, yerr)
                elif _yerr.ndim == 2:
                    _yerr = (yerr[0], yerr[1])
                else:
                    raise ValueError("Cannot interpret error array with shape %s" % (_yerr.shape,))
            except AttributeError:
                _xerr = (xerr, xerr)
    else:
        draw_central_value = True  # force drawing central value, if no y errors

    if _xerr is not None:
        if isinstance(_xerr, tuple) and len(_xerr)==1:
            _xerr = (xerr[0], xerr[0])
        elif isinstance(_xerr, tuple) and len(_xerr)==2:
            _xerr = (xerr[0], xerr[1])
        else:
            try:
                if _xerr.ndim == 1:
                    _xerr = (xerr, xerr)
                elif _xerr.ndim == 2:
                    _xerr = (xerr[0], xerr[1])
                else:
                    raise ValueError("Cannot interpret error array with shape %s" % (_xerr.shape,))
            except AttributeError:
                _xerr = (xerr, xerr)

    # -- do actual plotting

    if _xerr is not None:
        # calculate box corners for fill_between
        _x_lo = x - _xerr[0]
        _x_hi = x + _xerr[1]

        if continuous:
            # interleave x and y bin boundaries
            _x_plot = np.zeros(len(x)*2)
            _x_plot[0::2] = _x_lo
            _x_plot[1::2] = _x_hi
        else:
            _x_plot = np.array([_x_lo, _x_hi])
    else:
        raise ValueError('xerr cannot be None')

    if draw_central_value:
        if continuous:
            _y_plot_central = np.zeros(len(y) * 2)
            _y_plot_central[0::2] = y
            _y_plot_central[1::2] = y
        else:
            _y_plot_central = np.array([y, y])
        _modkwargs = deepcopy(kwargs)
        _modkwargs.pop('edgecolor', None)
        _modkwargs.pop('alpha', None)
        p_cv = axes.plot(_x_plot, _y_plot_central, **_modkwargs)
    else:
        p_cv = None

    if _yerr is not None:
        _y_lo = y - _yerr[0]
        _y_hi = y + _yerr[1]

        _y_plot_lo = np.zeros(len(y)*2)
        _y_plot_lo[0::2] = _y_lo
        _y_plot_lo[1::2] = _y_lo
        _y_plot_hi = np.zeros(len(y)*2)
        _y_plot_hi[0::2] = _y_hi
        _y_plot_hi[1::2] = _y_hi

        p_fb = axes.fill_between(_x_plot, _y_plot_lo, _y_plot_hi, **kwargs)
    else:
        p_fb = None

    # FIXME: (for _yerr!=None, return both artists somehow)
    if p_cv is not None and p_fb is not None:
        #return (p_fb, p_cv)  # doesn't work!
        return p_fb
    if p_cv is None:
        return p_fb
    if p_fb is None:
        #return p_cv  # doesn't work!
        return p_cv[0]



import numpy as np
import unittest

from kafe2.core.fitters.nexus import (
    NodeBase,
    Parameter, Alias,
    Function, Empty,
    Fallback, Tuple, Array, RootNode,

    NodeChildrenPrinter, NodeCycleChecker,

    _OPERATORS, _UNARY_OPERATORS,

    Nexus
)


class TestNodes(unittest.TestCase):

    DEFAULT_CONSTRUCTIBLE_NODE_TYPES = (
        NodeBase, Empty, RootNode
    )
    PARAMETRIC_NODE_TYPES = (
        Parameter, Alias, Function,
        Fallback, Tuple
    )

    @staticmethod
    def sum_function(a, b):
        return a + b

    def setUp(self):
        self.par_a = Parameter(3)
        self.par_b = Parameter(7)

        self.func_sum_a_b = Function(
            TestNodes.sum_function,
            parameters=(self.par_a, self.par_b)
        )

        self.empty_par = Empty()

        self.counter = 0
        self.sum = 0

        self.array_content = [0.1, 2.3, 4.5, 6.7, 8.9]

    # -- NodeBase

    def test_nodes_default_constructor(self):
        for _node_type in self.DEFAULT_CONSTRUCTIBLE_NODE_TYPES:
            # NodeBase constructor cannot be called because NodeBase has an abstract method.
            if _node_type is not NodeBase:
                _node_type()
                str(_node_type())

    def test_nodes_parametric_constructor(self):
        p = Parameter(None)
        str(p)

        a = Alias(Parameter(None))
        str(a)

        f = Function(lambda x: x, name='_lambda')
        str(f)

        fb = Fallback((f,))
        str(fb)

        t = Tuple([])
        str(t)

    def test_nodes_equality(self):
        p_1 = Parameter(None)
        p_2 = Parameter(None)
        self.assertEqual(p_1, p_1)
        self.assertNotEqual(p_1, p_2)

    def test_node_get_children_get_parents(self):
        par_1 = Parameter(3, name='par_1')
        par_2 = Parameter(2, name='par_2')
        par_lc_1 = par_1 + par_2
        par_lc_2 = par_1 - par_2

        for _par in par_1, par_2:
            self.assertEqual(
                _par.get_parents(),
                [par_lc_1, par_lc_2]
            )
        for _par in par_lc_1, par_lc_2:
            self.assertEqual(
                _par.get_children(),
                [par_1, par_2]
            )

    def test_node_iter_parents(self):
        par_1 = Parameter(3, name='par_1')
        par_2 = Parameter(2, name='par_2')
        par_lc_1 = par_1 + par_2
        par_lc_2 = par_1 - par_2

        for _par in par_1, par_2:
            self.assertEqual(
                [_parent for _parent in _par.iter_parents()],
                [par_lc_1, par_lc_2]
            )

    def test_node_add_child(self):
        par = Parameter(3, name='par')
        par.add_child(1)
        par.add_child(2)

        self.assertEqual(
            [p.value for p in par.get_children()],
            [1, 2]
        )

    def test_node_add_parent(self):
        par = Parameter(3, name='par')
        par_2 = Parameter(2, name='par_2')

        with self.assertRaises(TypeError):
            par.add_parent("notanode")

        # Manually adding parents is not allowed:
        with self.assertRaises(ValueError):
            par.add_parent(par_2)

        par_2.add_child(par)
        self.assertEqual(par.get_parents(), [par_2])

    def test_node_set_children(self):
        par = Parameter(3, name='par')
        child_1 = Parameter(1, name='child_1')
        child_2 = Parameter(2, name='child_2')

        par.set_children([child_1, child_2])

        self.assertEqual(
            par.get_children(),
            [child_1, child_2]
        )

    def test_node_remove_child(self):
        par = Parameter(3, name='par')
        child_1 = Parameter(1, name='child_1')
        child_2 = Parameter(2, name='child_2')

        par.set_children([child_1, child_2])
        par.remove_child(child_1)

        self.assertEqual(
            par.get_children(),
            [child_2]
        )

        with self.assertRaises(TypeError):
            par.remove_child("notanode")

    def test_node_remove_parent(self):
        par = Parameter(3, name='par')
        parent_1 = Parameter(1, name='parent_1')
        parent_2 = Parameter(2, name='parent_2')

        parent_1.add_child(par)
        parent_2.add_child(par)
        with self.assertRaises(ValueError):
            par.remove_parent(parent_1)
        self.assertEqual(
            par.get_parents(),
            [parent_1, parent_2]
        )
        parent_1.remove_child(par)
        self.assertEqual(
            par.get_parents(),
            [parent_2]
        )

        with self.assertRaises(TypeError):
            par.remove_parent("notanode")

    def test_parameter_replace(self):
        par = Parameter(3, name='par')
        selector = Function(
            lambda *args: args[1],
            name='selector',
            parameters=(
                Parameter('bla', name='first'),
                par,
                Parameter('blup', name='third'),
            ),
        )
        context = Parameter(2, name='pre_factor') * selector

        # store parents and children for comparison
        _children_before = par.get_children()
        _parents_before = par.get_parents()

        # replace node
        par_new = Parameter(7, name='par_new')
        par.replace(par_new)

        par_new.value = 6

        self.assertEqual(context.value, 12)

        # retrieve new parents and children for comparison
        _children_after = par_new.get_children()
        _parents_after = par_new.get_parents()

        self.assertEqual(_children_before, _children_after)
        self.assertEqual(_parents_before, _parents_after)

    def test_parameter_replace_literal(self):
        par = Parameter(3, name='par')
        selector = Function(
            lambda *args: args[1],
            name='selector',
            parameters=(
                Parameter('bla', name='first'),
                par,
                Parameter('blup', name='third'),
            ),
        )
        context = Parameter(2, name='pre_factor') * selector

        # replace node
        par.replace(7)

        self.assertEqual(context.value, 14)

    def test_function_replace(self):
        par = Parameter(3, name='par')
        selector = Function(
            lambda *args: args[1],
            name='selector',
            parameters=(
                Parameter('bla', name='first'),
                par,
                Parameter('blup', name='third'),
            ),
        )
        context = Parameter(2, name='pre_factor') * selector

        # store parents and children for comparison
        _children_before = selector.get_children()
        _parameters_before = selector.parameters
        _parents_before = selector.get_parents()

        # replace node
        selector_new = Function(lambda *args: args[2], name='selector_new')
        selector.replace(selector_new, other_children=False)

        # retrieve new parents and children for comparison
        _children_after = selector_new.get_children()
        _parents_after = selector_new.get_parents()
        _parameters_after = selector_new.parameters

        self.assertEqual(_children_before, _children_after)
        self.assertEqual(_parents_before, _parents_after)
        self.assertEqual(_parameters_before, _parameters_after)

        self.assertEqual(selector_new.value, 'blup')

    def test_replace_child(self):
        par_a = Parameter(4)
        par_b = Parameter(5)
        par_c = Parameter(6)
        par_a.add_child(par_b)
        self.assertEqual(par_a.get_children(), [par_b])
        with self.assertRaises(ValueError):
            par_a.replace_child(current_child=par_c, new_child=par_b)
        par_a.replace_child(current_child=par_b, new_child=par_c)
        with self.assertRaises(ValueError):
            par_a.replace_child(current_child=par_b, new_child=par_c)
        self.assertEqual(par_a.get_children(), [par_c])
        par_a.replace_child(current_child=par_c, new_child=7)
        self.assertNotEqual(par_a.get_children(), [par_b])
        self.assertNotEqual(par_a.get_children(), [par_c])
        self.assertIs(type(par_a.get_children()[0]), Parameter)

    # -- RootNode

    def test_root_node_replace(self):
        r = RootNode()
        with self.assertRaises(TypeError):
            r.replace(7)

    def test_root_node_parents(self):
        r = RootNode()
        with self.assertRaises(TypeError):
            r.add_parent(2)
        self.assertEqual([p for p in r.iter_parents()], [])
        self.assertEqual(r.get_parents(), [])

    # -- Empty

    def test_empty_value(self):
        e = Empty()
        with self.assertRaises(TypeError):
            _ = e.value
        with self.assertRaises(TypeError):
            e.value = 33

    # -- Parameter

    def test_parameter_get_value(self):
        par = Parameter(3)
        self.assertEqual(par.value, 3)

    def test_parameter_set_value(self):
        par = Parameter(3)
        par.value = 7
        self.assertEqual(par.value, 7)

    def test_parameter_name(self):
        par = Parameter(3, name='bla')
        self.assertEqual(par.name, 'bla')

    def test_parameter_invalid_name_raise(self):
        with self.assertRaises(ValueError):
            par = Parameter(3, name='2bla')

    def test_parameter_reserved_name_raise(self):
        for _rn in NodeBase.RESERVED_PARAMETER_NAMES:
            with self.assertRaises(ValueError):
                par = Parameter(3, name=_rn)

    def test_tuple_iter_raise(self):
        par = Parameter(3)
        with self.assertRaises(TypeError):
            iter(par)

    def test_parameter_binary_operation(self):
        par_a = Parameter(3)
        par_b = Parameter(7)
        for _op_name, (_op, _) in _OPERATORS.items():
            par_expr = _op(par_a, par_b)
            self.assertEqual(
                par_expr.value,
                _op(par_a.value, par_b.value)
            )

    def test_parameter_binary_operation_with_literal(self):
        par_a = Parameter(3)
        for _op_name, (_op, _) in _OPERATORS.items():
            par_expr = _op(par_a, 7)
            self.assertEqual(
                par_expr.value,
                _op(par_a.value, 7)
            )
            # reversed order
            par_expr = _op(7, par_a)
            self.assertEqual(
                par_expr.value,
                _op(7, par_a.value)
            )

    def test_parameter_unary_operation(self):
        par = Parameter(3)

        self.assertEqual(
            (+par).value, par.value)
        self.assertEqual(
            (-par).value, -(par.value))
        self.assertEqual(
            (~par).value, ~(par.value))

    # -- Alias

    def test_alias_value(self):
        par = Parameter(3)
        alias = Alias(par, name='alias')
        self.assertEqual(alias.value, par.value)
        with self.assertRaises(TypeError):
            alias.value = 5

    def test_multiple_alias_value(self):
        par = Parameter(3)
        alias = Alias(Alias(Alias(par)))
        self.assertEqual(alias.value, par.value)

    def test_alias_ref(self):
        par = Parameter(3)
        alias = Alias(par, name='alias')
        self.assertEqual(alias.ref, par)

    # -- Function

    def test_function_value(self):
        par_a = Parameter(3)
        par_b = Parameter(7)

        def func(a, b):
            return a*10 + b

        func_a_b = Function(func, parameters=(par_a, par_b))

        self.assertEqual(
            func_a_b.value,
            func(par_a.value, par_b.value),
        )

        # test value update
        par_a.value = -3

        self.assertEqual(
            func_a_b.value,
            func(par_a.value, par_b.value),
        )

        with self.assertRaises(TypeError):
            func_a_b.value = 5

    def test_function_value_update_frozen(self):
        par_a = Parameter(3)
        par_b = Parameter(7)

        def func(a, b):
            return a*10 + b

        func_a_b = Function(func, parameters=(par_a, par_b))

        self.assertEqual(
            func_a_b.value,
            func(par_a.value, par_b.value),
        )

        # test value update
        func_a_b.freeze()
        par_a.value = -3

        self.assertEqual(
            func_a_b.value,
            func(3, 7),
        )

        func_a_b.unfreeze()
        self.assertEqual(
            func_a_b.value,
            func(par_a.value, par_b.value),
        )

    def test_function_value_lambda(self):
        par_a = Parameter(3)
        par_b = Parameter(7)

        func_a_b = Function(
            lambda a, b: a*10 + b,
            name='_lambda',
            parameters=(par_a, par_b)
        )

        self.assertEqual(
            func_a_b.value,
            func_a_b.func(par_a.value, par_b.value),
        )

    def test_function_parameters(self):
        par_a = Parameter(3)
        par_b = Parameter(7)

        def func(a, b):
            return a*10 + b

        func_a_b = Function(func, parameters=(par_a, par_b))

        self.assertEqual(
            func_a_b.get_children(),
            [par_a, par_b],
        )
        self.assertEqual(
            func_a_b.parameters,
            [par_a, par_b],
        )

    def test_function_auto_parameters(self):
        def func(a, b):
            return a*10 + b

        func_a_b = Function(func)

        par_a = Parameter(3)
        par_b = Parameter(7)

        func_a_b.parameters = [par_a, par_b]

        self.assertEqual(
            func_a_b.parameters,
            [par_a, par_b],
        )
        self.assertEqual(func_a_b.value, 37)

    def test_function_auto_parameters_add_one_at_a_time(self):
        def func(a, b):
            return a*10 + b

        func_a_b = Function(func)

        par_a = Parameter(3)
        par_b = Parameter(7)

        func_a_b.add_parameter(par_a)
        func_a_b.add_parameter(par_b)

        self.assertEqual(
            func_a_b.parameters,
            [par_a, par_b],
        )
        self.assertEqual(func_a_b.value, 37)

    # -- Fallback

    def test_fallback(self):
        par_a = Parameter(14)
        par_b = Parameter(2)

        div_a_b = Fallback(
            (par_a / par_b, 'DIV0'),
            exception_type=ZeroDivisionError
        )

        # no exception yet
        self.assertEqual(
            div_a_b.value,
            par_a.value / par_b.value,
        )

        # make denominator zero
        par_b.value = 0

        # exception matches type -> fallback
        self.assertEqual(
            div_a_b.value,
            'DIV0',
        )

        # exception does not match type -> raise
        with self.assertRaises(ZeroDivisionError):
            Fallback(
                (par_a / par_b, 'DIV0'),
                exception_type=TypeError
            ).value

    def test_fallback_no_good_alternative(self):
        par_a = Parameter(14)

        div_a_b = Fallback(
            (par_a / 0, 2 * par_a / 0),
            exception_type=ZeroDivisionError
        )

        # no exception yet
        with self.assertRaises(RuntimeError):
            div_a_b.value,

    # -- Tuple

    def test_tuple_value(self):
        tuple_a = Tuple((14, 2))
        self.assertIs(type(tuple_a.value), tuple)
        self.assertEqual(tuple_a.value, (14, 2))

    def test_tuple_len(self):
        tuple_a = Tuple((14, 2))
        self.assertEqual(len(tuple_a), 2)

    def test_tuple_getitem(self):
        tuple_a = Tuple((14, 2))
        self.assertEqual(tuple_a.value, (14, 2))
        self.assertEqual(tuple_a[0].value, 14)
        self.assertEqual(tuple_a[1].value, 2)

    def test_tuple_setitem(self):
        tuple_a = Tuple((14, 2))
        tuple_a[0] = Parameter(15)
        tuple_a[1] = 1
        self.assertEqual(tuple_a.value, (15, 1))
        self.assertEqual(tuple_a[0].value, 15)
        self.assertEqual(tuple_a[1].value, 1)

    def test_tuple_nodes(self):
        tuple_a = Tuple((14, 2))
        tuple_a.nodes = (9, 4)
        self.assertEqual(tuple_a.value, (9, 4))

    def test_tuple_iter(self):
        tuple_a = Tuple((14, 2))
        self.assertEqual(
            tuple_a.nodes,
            [node for node in tuple_a]
        )

    def test_tuple_iter_values(self):
        tuple_a = Tuple((14, 2))
        self.assertEqual(
            tuple_a.value,
            tuple([value for value in tuple_a.iter_values()])
        )

    # -- Array

    def test_array_value(self):
        array_a = Array(self.array_content)
        self.assertIs(type(array_a.value), np.ndarray)
        self.assertTrue(np.all(array_a.value == self.array_content))

    def test_array_getitem(self):
        array_a = Array(self.array_content)
        for i, expected_value_i in enumerate(self.array_content):
            self.assertEqual(array_a[i].value, expected_value_i)

    def test_array_setitem(self):
        array_a = Array(self.array_content)
        array_a[3] = 0
        self.assertIs(type(array_a[3]), Parameter)
        self.assertTrue(np.all(array_a.value == [0.1, 2.3, 4.5, 0.0, 8.9]))

    def test_array_iter_values(self):
        array_a = Array(self.array_content)
        self.assertTrue(np.all(
            array_a.value == np.array([value for value in array_a.iter_values()])
        ))


class TestNodeVisitors(unittest.TestCase):

    def setUp(self):
        self.root = RootNode()
        self.empty = Empty(name="empty")
        self.par_a = Parameter(2, name="par_a")
        self.par_b = Parameter(3, name="par_b")
        self.par_c = Parameter(4, name="par_c")
        self.func_1 = self.par_a + self.par_b
        self.func_2 = self.par_a * self.par_c
        self.func_3 = Function(lambda a, b: a + b)
        self.alias = Alias(self.par_a, name="alias")
        self.par_a.set_children([self.par_b])
        self.root.set_children([self.empty, self.func_1, self.func_2, self.func_3, self.alias])

    # -- NodeChildrenPrinter

    def test_node_children_printer(self):
        self.root.print_descendants()

    # -- NodeCycleChecker

    def test_node_cycle_checker(self):

        NodeCycleChecker(self.par_a).run()

        self.par_b.set_children([self.par_a])

        with self.assertRaises(ValueError):
            NodeCycleChecker(self.par_a).run()


class TestNexus(unittest.TestCase):

    def setUp(self):
        self._nexus = Nexus()

    def test_add_get(self):
        par = Parameter('my_value')
        self._nexus.add(par)

        self.assertIs(
            self._nexus.get(par.name),
            par,
        )

    def test_add_get_inexistent(self):
        self.assertIs(
            self._nexus.get('bogus_name'),
            None,
        )

    def test_add_get_explicit_name(self):
        par = Parameter('my_value', name='par')
        self._nexus.add(par)

        self.assertIs(
            self._nexus.get('par'),
            par,
        )

    def test_add_existing_fail(self):
        par = Parameter('my_value')
        self._nexus.add(par)

        with self.assertRaises(ValueError):
            self._nexus.add(par, existing_behavior='fail')

    def test_add_existing_replace(self):
        par_orig = Parameter('my_original_value', name='par')
        self._nexus.add(par_orig)

        par_new = Parameter('my_new_value', name='par')
        self._nexus.add(par_new, existing_behavior='replace')

        self.assertIs(
            self._nexus.get('par'),
            par_new,
        )

    def test_add_existing_replace_alias(self):
        par_orig = Parameter('my_original_value', name='par')
        self._nexus.add(par_orig)
        alias_orig = Alias(par_orig, name='alias')
        self._nexus.add(alias_orig)

        par_new = Parameter('my_new_value', name='par')
        with self.assertRaises(ValueError):
            self._nexus.add(par_new, existing_behavior='replace_if_alias')
        self.assertIs(
            self._nexus.get('par'),
            par_orig,
        )
        alias_new = Parameter('my_new_value', name='alias')
        self._nexus.add(alias_new, existing_behavior='replace_if_alias')
        self.assertIs(
            self._nexus.get('alias'),
            alias_new,
        )

    def test_add_existing_replace_empty(self):
        par_orig = Parameter('my_original_value', name='par')
        self._nexus.add(par_orig)
        empty_orig = Empty('empty')
        self._nexus.add(empty_orig)

        par_new = Parameter('my_new_value', name='par')
        with self.assertRaises(ValueError):
            self._nexus.add(par_new, existing_behavior='replace_if_empty')
        self.assertIs(
            self._nexus.get('par'),
            par_orig,
        )
        empty_new = Parameter('my_new_value', name='empty')
        self._nexus.add(empty_new, existing_behavior='replace_if_empty')
        self.assertIs(
            self._nexus.get('empty'),
            empty_new,
        )

    def test_add_existing_ignore(self):
        par_orig = Parameter('my_original_value', name='par')
        self._nexus.add(par_orig)

        par_new = Parameter('my_new_value', name='par')
        self._nexus.add(par_new, existing_behavior='ignore')

        self.assertIs(
            self._nexus.get('par'),
            par_orig,
        )

    def test_add_existing_bogus(self):
        par = Parameter(2)
        self._nexus.add(par)
        with self.assertRaises(ValueError):
            self._nexus.add(par, existing_behavior='bogus')

    def test_add_expression_compare_value(self):
        a = Parameter(3)
        b = Parameter(5)

        expr = (a + b)
        expr.name = 'sum'

        self._nexus.add(a)
        self._nexus.add(b)
        self._nexus.add(expr)

        self.assertEqual(self._nexus.get('sum').value, 8)

    def test_add_expression_test_dependents_added(self):
        a = Parameter(3)
        b = Parameter(5)

        expr = (a - b) / (a + b)
        expr.name = 'asymm'

        expr_value = (a.value - b.value) / (a.value + b.value)

        self._nexus.add(expr)

        self.assertIs(self._nexus.get(a.name), a)
        self.assertIs(self._nexus.get(b.name), b)

    def test_add_function(self):
        def my_func(a, b=2):
            return 2 * a + b

        func_node = self._nexus.add_function(
            func=my_func
        )

        # check if name is as in signature
        self.assertEqual(func_node.name, 'my_func')

        # check if parameter names are as in signature
        self.assertEqual(
            [p.name for p in func_node.parameters],
            ['a', 'b']
        )

        self.assertTrue(isinstance(self._nexus.get('a'), Empty))
        self.assertTrue(isinstance(self._nexus.get('b'), Parameter))
        self.assertEqual(self._nexus.get('b').value, 2)

    def test_add_function_existing_parameters(self):

        a = self._nexus.add(Parameter(4, name='a'))
        b = self._nexus.add(Parameter(5, name='b'))

        def my_func(a, b):
            return 2 * a + b

        func_node = self._nexus.add_function(
            func=my_func
        )

        # check if name is as in signature
        self.assertEqual(func_node.name, 'my_func')

        # check if parameter values correspond to existing parameters
        self.assertEqual(
            [p.value for p in func_node.parameters],
            [a.value, b.value]
        )
        # check function value
        self.assertEqual(func_node.value, my_func(a.value, b.value))

    def test_add_function_existing_parameter_with_default(self):

        a = self._nexus.add(Parameter(4, name='a'))
        b = self._nexus.add(Parameter(9, name='b'))

        def my_func(a=4, b=200):
            return 2 * a + b

        with self.assertWarns(UserWarning) as w:
            func_node = self._nexus.add_function(
                func=my_func
            )

        # check if parameter values correspond to existing parameters
        self.assertEqual(
            [p.value for p in func_node.parameters],
            [a.value, b.value]
        )

        self.assertIn('Ignoring default value', w.warning.args[0])
        self.assertIn('conflicting value', w.warning.args[0])

    def test_add_function_custom_name(self):
        def my_func(a, b):
            return 2 * a + b

        func_node = self._nexus.add_function(
            func=my_func,
            func_name='new_func_name',
        )

        # check if name is as in signature
        self.assertEqual(func_node.name, 'new_func_name')

    def test_add_function_custom_par_names(self):
        x = self._nexus.add(Parameter(7, name='x'))
        y = self._nexus.add(Parameter(8, name='y'))

        def my_func(a, b):
            return 2 * a + b

        func_node = self._nexus.add_function(
            func=my_func,
            par_names=['x', 'y']
        )

        # check function value
        self.assertEqual(func_node.value, my_func(x.value, y.value))

    def test_add_function_varargs(self):
        x = self._nexus.add(Parameter(7, name='x'))
        y = self._nexus.add(Parameter(8, name='y'))

        def my_func(*args):
            return sum(args)

        func_node = self._nexus.add_function(
            func=my_func,
            par_names=['x', 'y']
        )

        # check function value
        self.assertEqual(func_node.value, my_func(x.value, y.value))

    def test_add_function_var_kwargs(self):
        def my_func(a, b, c, **kwargs):
            return a + b + c + np.sum(kwargs.values())

        with self.assertRaises(ValueError):
            self._nexus.add_function(func=my_func)

    def test_add_function_combine_defaults(self):
        def my_func_1(a, b=2):
            return a + b

        def my_func_2(b, a=3):
            return a * b

        func_node_1 = self._nexus.add_function(my_func_1)
        with self.assertRaises(TypeError):
            _ = func_node_1.value
        func_node_2 = self._nexus.add_function(my_func_2)
        # Parameters of second function are reversed:
        self.assertEqual(func_node_1.get_children(), func_node_2.get_children()[::-1])
        self.assertEqual(func_node_1.parameters, func_node_2.parameters[::-1])
        self.assertEqual(func_node_1.value, 5)
        self.assertEqual(func_node_2.value, 6)

    def test_add_alias(self):
        with self.assertRaises(ValueError):
            self._nexus.add_alias(name="alias_1", alias_for="par")
        par = Parameter(6, name="par")
        self._nexus.add(par)
        self._nexus.add_alias(name="alias_1", alias_for="par")
        self.assertIs(type(self._nexus.get("alias_1")), Alias)
        self.assertEqual(self._nexus.get("alias_1").value, 6)
        alias_2 = Alias(ref=par, name="alias_2")
        self._nexus.add(alias_2)
        self.assertIs(type(self._nexus.get("alias_2")), Alias)
        self.assertEqual(self._nexus.get("alias_2").value, 6)
        with self.assertRaises(ValueError):
            self._nexus.add_alias(name="alias_3", alias_for="DEADBEEF")

    def test_add_detect_cycles(self):
        # define pair of infinitely recursive functions
        def my_rec_func(my_rec_func_2):
            return 2 * my_rec_func_2

        def my_rec_func_2(my_rec_func):
            return 2 * my_rec_func

        func_node = self._nexus.add_function(
            func=my_rec_func,
            par_names=['my_rec_func_2']
        )

        # adding second one triggers cycle detection
        with self.assertRaises(ValueError):
            self._nexus.add_function(
                func=my_rec_func_2,
                existing_behavior='replace',
            )
            _ = func_node.value

    def test_add_dependency(self):
        def test_func(x=3):
            return 2 * x

        func_node = self._nexus.add_function(
            func=test_func
        )

        y = self._nexus.add(
            Parameter(4, name='y')
        )

        self.assertEqual(func_node.value, 6)
        self.assertEqual(func_node.stale, False)

        # here 'test_func' does not go stale on 'y' update
        y.value = 8
        self.assertEqual(func_node.stale, False)

        self._nexus.add_dependency('test_func', depends_on='y')

        # now 'test_func' depends on 'y' and should go stale
        y.value = 23
        self.assertEqual(func_node.stale, True)
        self.assertEqual(func_node.value, 6)

        with self.assertRaises(ValueError):
            self._nexus.add_dependency('DEADBEEF', depends_on='y')
        with self.assertRaises(ValueError):
            self._nexus.add_dependency('y', depends_on=['a', 'b'])

    def test_get_value_dict(self):
        self._nexus.add(Parameter(1, name="a"))
        self._nexus.add(Parameter(2, name="b"))
        self._nexus.add(Parameter(3, name="c"))
        self._nexus.add_function(lambda a, b, c: a + b * c, func_name="func")
        value_dict = self._nexus.get_value_dict()
        self.assertEqual(len(value_dict), 4)
        self.assertEqual(value_dict["a"], 1)
        self.assertEqual(value_dict["b"], 2)
        self.assertEqual(value_dict["c"], 3)
        self.assertEqual(value_dict["func"], 7)
        value_dict = self._nexus.get_value_dict(node_names=["b", "a", "func"])
        self.assertEqual(len(value_dict), 3)
        self.assertEqual(value_dict["a"], 1)
        self.assertEqual(value_dict["b"], 2)
        self.assertEqual(value_dict["func"], 7)
        with self.assertRaises(ValueError):
            _ = self._nexus.get_value_dict(error_behavior="bogus")

    def test_print_state(self):
        self._nexus.add(Parameter(1, name="a"))
        self._nexus.add(Parameter(2, name="b"))
        self._nexus.add(Parameter(3, name="c"))
        self._nexus.add_function(lambda a, b, c: a + b * c, func_name="func")
        self._nexus.print_state()



from __future__ import print_function

import numpy as np
from scipy.special import gammainccinv, gammaincc

__all__ = ['ConfidenceLevel']


class ConfidenceLevel(object):
    """
    Helper class for handling the conversion from confidence levels to sigma values and vice versa.
    """
    def __init__(self, n_dimensions, cl=None, sigma=None):
        self.ndim = n_dimensions
        if (cl is None and sigma is None) or (cl is not None and sigma is not None):
            raise ValueError("Exactly one out of cl and sigma must be defined!")
        if cl is not None:
            self.cl = cl
        if sigma is not None:
            self.sigma = sigma

    def __str__(self):
        return "<ConfidenceLevel (d=%d): %.4g%% (%.3g-sigma)>" % (
            self.ndim, self.cl*100., self.sigma)

    def _calc_sigma_from_cl(self):
        self._sigma = np.sqrt(2 * gammainccinv(self.ndim/2.0, 1.0 - self.cl))

    def _calc_cl_from_sigma(self):
        self._cl = 1. - gammaincc(self.ndim/2.0, self.sigma**2/2.0)

    @property
    def cl(self):
        if self._cl is None:
            self._calc_cl_from_sigma()
        return self._cl

    @cl.setter
    def cl(self, new_cl):
        if new_cl <= 0 or new_cl >= 1:
            raise ValueError(
                "Confidence level must be greater than 0 and less than 1. Got: %g" % (new_cl,))
        self._cl = float(new_cl)
        self._sigma = None

    @property
    def sigma(self):
        if self._sigma is None:
            self._calc_sigma_from_cl()
        return self._sigma

    @sigma.setter
    def sigma(self, new_sigma):
        if new_sigma <= 0:
            raise ValueError("Sigma value must be greater than 0! Got: %g" % (new_sigma,))
        self._sigma = float(new_sigma)
        self._cl = None

    @property
    def ndim(self):
        return self._ndim

    @ndim.setter
    def ndim(self, new_ndim):
        if not isinstance(new_ndim, int):
            raise ValueError(
                "Number of dimensions must be of type int! Received type: %s" % type(new_ndim))
        if new_ndim <= 0:
            raise ValueError(
                "Number of dimensions must be greater 0! Received: %d" % (new_ndim,)
            )
        self._ndim = new_ndim

    @property
    def sigma_string(self):
        return "%.3g-sigma" % (self.sigma,)

    @property
    def sigma_latex_string(self):
        return r"%g$\sigma$" % (self.sigma,)

    @property
    def cl_string(self):
        return "%.4g%% CL" % (self.cl*100,)

    @property
    def cl_latex_string(self):
        return r"$%.4g\%%$ CL" % (self.cl*100,)



"""This submodule contains objects for handling different representations of model function formatters and argument
formatters.

:synopsis: This submodule contains objects for handling different representations of model function
        formatters and argument formatters.

.. moduleauthor:: Johannes Gaessler <johannes.gaessler@student.kit.edu>
"""

from ._base import *
from .yaml_drepr import *



import abc
import unittest
import numpy as np
import six

from scipy import stats

from kafe2.core.minimizers import AVAILABLE_MINIMIZERS

from kafe2.config import kc

from kafe2.fit import HistFit, HistContainer
from kafe2.fit.histogram.cost import HistCostFunction_NegLogLikelihood

from kafe2.test.fit.test_fit import AbstractTestFit


def simple_chi2(data, model):
    return np.sum((data - model)**2)


def hist_model_density(x, mu=14., sigma=3.):
    return stats.norm(mu, sigma).pdf(x)

def hist_model_density_antideriv(x, mu=14., sigma=3.):
    return stats.norm(mu, sigma).cdf(x)



class TestHistFitBasicInterface(AbstractTestFit, unittest.TestCase):

    MINIMIZER = 'scipy'

    def setUp(self):
        self._ref_n_bins = 11
        self._ref_n_bin_range = (-3, 25)
        self._ref_bin_edges = np.linspace(self._ref_n_bin_range[0], self._ref_n_bin_range[1], self._ref_n_bins+1)

        self._ref_entries = np.array([11.47195963,   9.96715403,  19.90275216,  13.65225802,
                                      18.52670233,  17.79707059,  11.5441954 ,  18.42331074,
                                      13.3808496 ,  18.40632518,  13.21694177,  15.34569261,
                                       9.85164713,  11.56275047,  12.42687109,   9.43924719,
                                      15.03632673,  10.76203991,  15.16788754,  13.78683534,
                                      10.84686619,  17.29655143,  12.58404203,  13.10171453,
                                      20.23773075,  15.00936796,  15.38972301,  12.94254624,
                                      19.01920031,  19.67747312,  14.53196033,  13.78898329,
                                      17.34040707,  17.25328064,  13.64052295,  11.80461841,
                                      14.21872985,  11.59054406,  20.28243216,  15.43544627,
                                      15.85964469,  13.68948162,  14.55574162,  15.70198922,
                                      12.85548112,  14.05419407,  15.85172907,  19.2951937 ,
                                      12.70652203,  13.47822913,  12.84301792,  13.65407699,
                                      12.81908235,  12.73065184,  23.34917367,  11.38895625,
                                       9.41615435,   7.24554349,  11.4112971 ,  14.54736404,
                                      15.84186607,  15.33880379,   9.93177073,  15.14220967,
                                      15.19535013,  11.00033758,  11.83671083,  13.02411982,
                                      18.43323913,  12.06272124,  20.57549426,  12.97405382,
                                      13.35969913,  10.76072424,  16.35912216,  14.52842827,
                                      16.97221766,  13.02791932,  14.36097113,  16.93839498,
                                      15.60091018,  11.55044489,  14.06915886,  17.64985576,
                                      11.59865691,  13.41486068,  14.25999508,  10.70887047,
                                       4.08280244,  13.1043861 ,  14.62321312,  14.85894591,
                                      14.89235398,  10.60967181,  15.22310211,  10.77853626,
                                      14.56823312,  14.46093346,  13.34031129,  14.14203599])

        self._ref_hist_cont = HistContainer(
            self._ref_n_bins, self._ref_n_bin_range, bin_edges=None, fill_data=self._ref_entries)


        # reference initial values
        self._ref_initial_pars = np.array([14., 3.])
        self._ref_initial_model = (
            hist_model_density_antideriv(self._ref_bin_edges[1:], *self._ref_initial_pars) -
            hist_model_density_antideriv(self._ref_bin_edges[:-1], *self._ref_initial_pars)) * len(self._ref_entries)

        # fit data
        self._ref_data, _ = np.histogram(
            self._ref_entries,
            bins=self._ref_n_bins,
            range=self._ref_n_bin_range)

        # pre-fit cost value
        self._ref_initial_cost_nll = \
            -2*np.sum(stats.poisson.logpmf(self._ref_data, self._ref_initial_model))
        self._ref_initial_cost_chi2 = simple_chi2(self._ref_data, self._ref_initial_model)

        # reference fit result values
        self._nominal_fit_result_pars_nll = np.array([14.18427759, 3.02257722])
        self._nominal_fit_result_pars_chi2 = np.array([13.82779489, 2.62746457])

        self._nominal_fit_result_model_nll = (
            hist_model_density_antideriv(self._ref_bin_edges[1:], *self._nominal_fit_result_pars_nll) -
            hist_model_density_antideriv(self._ref_bin_edges[:-1], *self._nominal_fit_result_pars_nll)) * len(self._ref_entries)
        self._nominal_fit_result_model_chi2 = (
            hist_model_density_antideriv(self._ref_bin_edges[1:], *self._nominal_fit_result_pars_chi2) -
            hist_model_density_antideriv(self._ref_bin_edges[:-1], *self._nominal_fit_result_pars_chi2)) * len(self._ref_entries)

        self._nominal_fit_result_cost_nll = \
            -2*np.sum(stats.poisson.logpmf(self._ref_data, self._nominal_fit_result_model_nll))
        self._nominal_fit_result_cost_chi2 = simple_chi2(
            self._ref_data, self._nominal_fit_result_model_chi2)

        # helper dict with all reference property values
        self._ref_prop_dict = dict(
            did_fit=False,
            model_count=1,

            parameter_values=self._ref_initial_pars,
            parameter_names=('mu', 'sigma'),

            data=self._ref_data,
            model=self._ref_initial_model,
        )

    def _get_fit(self, model_density_function=None, bin_evaluation="numerical", cost_function=None):
        '''convenience'''

        model_density_function = model_density_function or hist_model_density

        # TODO: fix default
        cost_function = cost_function or HistCostFunction_NegLogLikelihood(
            data_point_distribution='poisson')

        _fit = HistFit(
            data=self._ref_hist_cont,
            model_function=model_density_function,
            bin_evaluation=bin_evaluation,
            cost_function=cost_function,
            minimizer=self.MINIMIZER
        )
        _fit.add_error(1.0)  # only considered for chi2

        return _fit

    def _get_test_fits(self):
        return {
            # numeric integration takes too long for testing
            #'default': \
            #    self._get_fit(),
            'explicit_chi2': \
                self._get_fit(cost_function=simple_chi2, bin_evaluation=hist_model_density_antideriv),
            'model_with_antiderivative': \
                self._get_fit(bin_evaluation=hist_model_density_antideriv),
        }

    def test_initial_state(self):
        self.run_test_for_all_fits(
            dict(
                self._ref_prop_dict,
                cost_function_value=np.float64(self._ref_initial_cost_nll),  # TODO: fix type
            ),
            fit_names=['default', 'model_with_antiderivative']
        )
        self.run_test_for_all_fits(
            dict(
                self._ref_prop_dict,
                cost_function_value=np.float64(self._ref_initial_cost_chi2),  # TODO: fix type
            ),
            fit_names=['explicit_chi2']
        )

    def test_fit_results(self):
        self.run_test_for_all_fits(
            dict(
                self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars_nll,
                model=self._nominal_fit_result_model_nll,
                did_fit=True,
                cost_function_value=np.float64(self._nominal_fit_result_cost_nll),
            ),
            fit_names=['default', 'model_with_antiderivative'],
            call_before_fit=lambda f: f.do_fit(),
            rtol=1e-2
        )
        self.run_test_for_all_fits(
            dict(
                self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars_chi2,
                model=self._nominal_fit_result_model_chi2,
                did_fit=True,
                cost_function_value=np.float64(self._nominal_fit_result_cost_chi2),
            ),
            fit_names=['explicit_chi2'],
            call_before_fit=lambda f: f.do_fit(),
            rtol=1e-2
        )

    def test_update_data(self):
        _fit = self._get_fit(bin_evaluation=hist_model_density_antideriv)

        _new_entries = np.array([
            19.424357258680693, 19.759361803397155, 18.364336396273362, 18.36464562195573,
            21.12001399388072, 18.08232152646234, 20.881797998050466, 19.799071480564336,
            20.00149234521563, 18.45879580610016, 20.507568873033875, 20.30648149729429,
            20.877350386469818, 19.381958240681676, 19.64442840120083, 19.141355147035597,
            19.94101630498644, 20.504982848860507, 17.524033521076785, 20.73543156774036,
            21.163113940992737, 19.05979241568689, 20.196952015154917, 20.40130402635463,
            21.80417387186999, 20.611530513961164, 19.099481246155072, 21.817653009909904,
            19.75055842274471, 20.103815812928232, 18.174017677733538, 21.047249981725685,
            21.262823911834488, 21.536864685525888, 18.813610758324447, 21.499655806695163,
            19.933264932657465, 21.954933746841995, 17.78470283680212, 19.8917212343489,
            19.372012624773838, 18.9520723656944, 19.9905553737993, 18.22737716365809,
            22.208437406472243, 19.875706306083835, 19.17672889225326, 20.10750939196147,
            20.093938177045032, 19.857292210131092, 20.17843836897936, 20.58803422718744,
            19.936410829343984, 19.050688989087522, 18.46936492682146, 21.90955106395087,
            19.661176212242154, 22.2764766192496, 19.850200163818528, 18.49289303805954,
            19.7563960302135, 20.940311019530235, 19.12732791777932, 22.09224225083453,
            20.225667564052465, 20.10787811564912, 18.660130651239726, 18.356069221596094,
            20.278651217320608, 18.62176541545302, 18.747451690981315, 19.81307693501857,
            19.34065619310232, 19.56998674371285, 19.885577923257177, 18.81752399043877,
            20.67686318083984, 20.265021790145465, 19.982547007042093, 19.581967230877964,
            18.486722000426457, 19.83143305661045, 21.252124382516378, 20.152988937293436,
            18.917354464336892, 18.349803731030892, 21.32702043081207, 22.410955706069743,
            20.972404800516973, 19.615870251101295, 19.013627387925588, 19.54487437668081,
            20.538542465210206, 18.626198427902466, 20.221745437611307, 19.064952809088076])

        _fit.data = HistContainer(
            8,
            (17, 23),
            fill_data=_new_entries
        )
        _fit.add_error(err_val=1.0)

        _ref_data, _ = np.histogram(
            _new_entries,
            bins=8,
            range=(17, 23)
        )

        self._assert_fit_properties(
            _fit,
            dict(
                data=_ref_data,
            )
        )

        _fit.do_fit()

        self._assert_fit_properties(
            _fit,
            dict(
                data=_ref_data,
                parameter_values=np.array([19.83815938,  1.1729322]),
            ),
            rtol=1e-2
        )

    def test_reserved_parameter_names_raise(self):
        def dummy_model(x, data):
            pass

        with self.assertRaises(ValueError) as _exc:
            HistFit(data=self._ref_hist_cont,
                    model_function=dummy_model,
                    minimizer=self.MINIMIZER)

        self.assertIn('reserved', _exc.exception.args[0])
        self.assertIn('data', _exc.exception.args[0])

    def test_model_no_pars_raise(self):
        def dummy_model():
            pass

        with self.assertRaises(ValueError) as _exc:
            HistFit(data=self._ref_hist_cont,
                    model_function=dummy_model,
                    bin_evaluation=dummy_model,
                    minimizer=self.MINIMIZER)

        self.assertIn(
            "needs at least one parameter",
            _exc.exception.args[0])

    def test_model_no_pars_beside_x_raise(self):
        def dummy_model(x):
            pass

        with self.assertRaises(ValueError) as _exc:
            HistFit(data=self._ref_hist_cont,
                    model_function=dummy_model,
                    bin_evaluation=dummy_model,
                    minimizer=self.MINIMIZER)

        self.assertIn(
            "needs at least one parameter besides",
            _exc.exception.args[0])

    def test_model_varargs_raise(self):
        # TODO: raise even without 'par'
        def dummy_model(x, par, *varargs):
            pass

        with self.assertRaises(ValueError) as _exc:
            HistFit(data=self._ref_hist_cont,
                    model_function=dummy_model,
                    bin_evaluation=dummy_model,
                    minimizer=self.MINIMIZER)

        self.assertIn('variable', _exc.exception.args[0])
        self.assertIn('varargs', _exc.exception.args[0])

    def test_model_varkwargs_raise(self):
        # TODO: raise even without 'par'
        def dummy_model(x, par, **varkwargs):
            pass

        with self.assertRaises(ValueError) as _exc:
            HistFit(data=self._ref_hist_cont,
                    model_function=dummy_model,
                    bin_evaluation=dummy_model,
                    minimizer=self.MINIMIZER)

        self.assertIn('variable', _exc.exception.args[0])
        self.assertIn('varkwargs', _exc.exception.args[0])

    def test_model_varargs_varkwargs_raise(self):
        # TODO: raise even without 'par'
        def dummy_model(x, par, *varargs, **varkwargs):
            pass

        with self.assertRaises(ValueError) as _exc:
            HistFit(data=self._ref_hist_cont,
                    model_function=dummy_model,
                    bin_evaluation=dummy_model,
                    minimizer=self.MINIMIZER)

        self.assertIn('variable', _exc.exception.args[0])
        self.assertIn('varargs', _exc.exception.args[0])
        # TODO: enable when implemented
        #self.assertIn('varkwargs', _exc.exception.args[0])

    def test_model_and_antiderivative_different_signatures_raise(self):
        def dummy_model(x, mu, sigma):
            pass

        def dummy_model_antiderivative(x, mu, bogus):
            pass

        with self.assertRaises(ValueError) as _exc:
            HistFit(data=self._ref_hist_cont,
                    model_function=dummy_model,
                    bin_evaluation=dummy_model_antiderivative,
                    minimizer=self.MINIMIZER)

    def test_model_and_antiderivative_no_defaults(self):
        def legendre_grade_2(x, a=1, b=2, c=3):
            return a + b * x + c * 0.5 * (3 * x ** 2 - 1)

        def legendre_grade_2_integrated(x, a, b, c):
            return 0.5 * x * (2 * a + b * x + c * (x ** 2 - 1))

        # should not raise an error
        HistFit(data=self._ref_hist_cont,
                model_function=legendre_grade_2,
                bin_evaluation=legendre_grade_2_integrated,
                minimizer=self.MINIMIZER)

    def test_report_before_fit(self):
        # TODO: check report content
        _buffer = six.StringIO()
        _fit = self._get_fit(bin_evaluation=hist_model_density_antideriv)
        _fit.report(output_stream=_buffer)
        self.assertNotEqual(_buffer.getvalue(), "")

    def test_report_after_fit(self):
        # TODO: check report content
        _buffer = six.StringIO()
        _fit = self._get_fit(bin_evaluation=hist_model_density_antideriv)
        _fit.do_fit()
        _fit.report(output_stream=_buffer)
        self.assertNotEqual(_buffer.getvalue(), "")

    def test_fit_scalar_model_function(self):
        def _model_function(x, a):
            return a

        _container = HistContainer(n_bins=2, bin_range=(0, 2), fill_data=[0.5, 1.5, 1.5])
        _fit = HistFit(_container, _model_function, density=False)
        _fit.do_fit()
        self.assertAlmostEqual(_fit.parameter_values[0], 1.5, places=4)
        _fit.report()



import numpy as np
from scipy.stats import norm
import unittest
import six
from abc import ABCMeta

from kafe2 import CustomFit, HistContainer, HistFit, IndexedFit, MultiFit, XYFit
from kafe2.test.tools import calculate_expected_fit_parameters_xy
from kafe2.test.fit.test_fit import AbstractTestFit
from kafe2.fit.util.function_library import quadratic_model, quadratic_model_derivative

_cannot_import_IMinuit = False
try:
    from kafe2.core.minimizers.iminuit_minimizer import MinimizerIMinuit
except (ImportError, SyntaxError):
    _cannot_import_IMinuit = True


class TestSharedErrorLogic(AbstractTestFit, unittest.TestCase):

    def _get_multifit(
            self, hist_fit=True, xy_fit_1=True, indexed_fit=True, xy_fit_2=True, reverse=False):
        _fits = []
        if hist_fit:
            _hist_container = HistContainer(
                n_bins=10, bin_range=(-5, 5), fill_data=self._hist_data)
            _hist_fit = HistFit(_hist_container)
            _fits.append(_hist_fit)
        if xy_fit_1:
            _xy_fit_1 = XYFit(xy_data=[self._x_data, self._y_data_1])
            _xy_fit_1.add_error("y", 1.0)
            _fits.append(_xy_fit_1)
        if indexed_fit:
            _indexed_fit = IndexedFit(
                self._indexed_data, TestSharedErrorLogic.indexed_model_function)
            _indexed_fit.add_error(1.0)
            _fits.append(_indexed_fit)
        if xy_fit_2:
            _xy_fit_2 = XYFit(xy_data=[self._x_data, self._y_data_2])
            _xy_fit_2.add_error("y", 1.0)
            _fits.append(_xy_fit_2)
        if reverse:
            _fits = reversed(_fits)
        return MultiFit(_fits)

    @staticmethod
    def indexed_model_function(a, b):
        return np.arange(10) * a + b

    def setUp(self):
        np.random.seed(0)
        self._hist_data = np.random.normal(size=100)*2+1
        self._x_data = np.arange(10) + 0.1 * np.random.normal()
        self._y_data = 2.3 * self._x_data + 4.5 + np.random.normal(size=10)
        self._y_data_1 = self._y_data + np.random.normal(size=10)
        self._y_data_2 = self._y_data + np.random.normal(size=10)
        self._indexed_data = self._y_data + np.random.normal(size=10)

    def test_with_vs_without_hist_fit(self):
        _multifit_hist = self._get_multifit(hist_fit=True)
        _multifit_hist.add_error(err_val=1.0, fits=[1, 2, 3], axis="y")
        _multifit_hist.add_error(err_val=0.1, fits=[1, 3], axis="x")
        _multifit_hist.do_fit()

        _multifit_no_hist = self._get_multifit(hist_fit=False)
        _multifit_no_hist.add_error(err_val=1.0, fits=[0, 1, 2], axis="y")
        _multifit_no_hist.add_error(err_val=0.1, fits=[0, 2], axis="x")
        _multifit_no_hist.do_fit()

        self._assert_values_equal(
            "parameter_values", _multifit_hist.parameter_values[2:],
            _multifit_no_hist.parameter_values)
        self._assert_values_equal(
            "parameter_errors", _multifit_hist.parameter_errors[2:],
            _multifit_no_hist.parameter_errors)
        self._assert_values_equal(
            "parameter_cov_mat", _multifit_hist.parameter_cov_mat[2:, 2:],
            _multifit_no_hist.parameter_cov_mat)
        self.assertTrue(
            _multifit_hist.cost_function_value - _multifit_no_hist.cost_function_value > 10.0)

    def test_reversed_same_result(self):
        _multifit = self._get_multifit(reverse=False)
        _multifit.add_error(err_val=1.0, fits=[1, 2, 3], axis="y")
        _multifit.add_error(err_val=0.1, fits=[1, 3], axis="x")
        _multifit.do_fit()

        _multifit_reversed = self._get_multifit(reverse=True)
        _multifit_reversed.add_error(err_val=1.0, fits=[0, 1, 2], axis="y")
        _multifit_reversed.add_error(err_val=0.1, fits=[0, 2], axis="x")
        _multifit_reversed.do_fit()

        self._assert_fit_results_equal(
            _multifit, _multifit_reversed, fit_2_permutation=[2, 3, 0, 1])

    def test_different_error_order_same_result(self):
        _multifit = self._get_multifit()
        _multifit.add_error(err_val=1.0, fits=[1, 2, 3], axis="y")
        _multifit.add_error(err_val=0.1, fits=[1, 3], axis="x")
        _multifit.do_fit()

        _multifit_reordered = self._get_multifit()
        _multifit_reordered.add_error(err_val=0.1, fits=[1, 3], axis="x")
        _multifit_reordered.add_error(err_val=1.0, fits=[1, 2, 3], axis="y")
        _multifit_reordered.do_fit()


        self._assert_fit_results_equal(_multifit, _multifit_reordered)

    def test_permuted_error_indices_same_result(self):
        _multifit = self._get_multifit()
        _multifit.add_error(err_val=1.0, fits=[1, 2, 3], axis="y")
        _multifit.add_error(err_val=0.1, fits=[1, 3], axis="x")
        _multifit.do_fit()

        _multifit_permuted = self._get_multifit()
        _multifit_permuted.add_error(err_val=1.0, fits=[1, 3, 2], axis="y")
        _multifit_permuted.add_error(err_val=0.1, fits=[3, 1], axis="x")
        _multifit_permuted.do_fit()

        self._assert_fit_results_equal(_multifit, _multifit_permuted)

    def test_add_shared_error_raise(self):
        _multifit = self._get_multifit()
        with self.assertRaises(ValueError):
            _multifit.add_error(err_val=1.0, fits="all", axis="z")
        with self.assertRaises(ValueError):
            _multifit.add_error(err_val=1.0, fits=[0, 1])
        with self.assertRaises(ValueError):
            _multifit.add_error(err_val=1.0, fits=[1, 3], axis=None)
        with self.assertRaises(ValueError):
            _multifit.add_error(
                err_val=1.0, fits=[1, 3], axis="y", relative=True, reference="modata")
        with self.assertRaises(ValueError):
            _multifit.add_error(
                err_val=1.0, fits=[1, 3], axis="x", relative=True, reference="model")
        with self.assertRaises(ValueError):
            _multifit.add_error(
                err_val=1.0, fits=[1, 3], axis="y", relative=True, reference="model")
        with self.assertRaises(ValueError):
            _multifit.add_error(
                err_val=1.0, fits=[1, 3], axis="y", relative=True, reference="data")

    def test_different_starting_parameters(self):
        _multifit_default = self._get_multifit()
        _multifit_default.do_fit()
        _multifit = self._get_multifit()
        _multifit.set_parameter_values(**{'mu': 1.2, 'sigma': 0.8, 'a': 1.2, 'b': 0.8})
        _multifit.do_fit()
        self._assert_fit_results_equal(_multifit_default, _multifit)


@six.add_metaclass(ABCMeta)
class TestMultiFit(AbstractTestFit, unittest.TestCase):

    @staticmethod
    def _split_data(data, axis=0):
        _random = np.random.rand(data.shape[axis])
        _split_indices_1 = np.argwhere(_random < 0.5).flatten()
        _split_indices_2 = np.argwhere(_random >= 0.5).flatten()
        if axis == 0:
            return data[_split_indices_1], data[_split_indices_2]
        if axis == 1:
            return data[:, _split_indices_1], data[:, _split_indices_2]
        raise Exception()

    def _assert_fits_valid_and_equal(
            self, _fit_1, _fit_2, atol=5e-3, rtol=2e-4, check_cost_function_value=True):
        with self.subTest("parameter_names"):
            assert len(_fit_1.parameter_names) == len(_fit_2.parameter_names)
            for _par_name_1, _par_name_2 in zip(_fit_1.parameter_names, _fit_2.parameter_names):
                assert _par_name_1 == _par_name_2
        self._assert_fit_results_equal(
            _fit_1, _fit_2, rtol=rtol, atol=atol,
            check_cost_function_value=check_cost_function_value)

        _fit_1.do_fit()
        _fit_2.do_fit()

        with self.subTest("parameter_values changed during fit"):
            for _parameter_value in _fit_1.parameter_values:
                assert _parameter_value != 1.0
            for _parameter_value in _fit_2.parameter_values:
                assert _parameter_value != 1.0
        self._assert_fit_results_equal(
            _fit_1, _fit_2, rtol=rtol, atol=atol,
            check_cost_function_value=check_cost_function_value)

    def _assert_fits_valid_and_equal_double(
            self, regular_fit, double_fit, atol=5e-3, rtol=2e-4, check_cost_function_value=False):
        with self.subTest("parameter_names"):
            assert len(regular_fit.parameter_names) == len(double_fit.parameter_names)
            for _par_name_1, _par_name_2 in zip(regular_fit.parameter_names, double_fit.parameter_names):
                assert _par_name_1 == _par_name_2
        self._assert_fit_results_equal(
            regular_fit, double_fit, rtol=rtol, atol=atol,
            check_cost_function_value=check_cost_function_value, fit_2_is_double_fit=True)

        regular_fit.do_fit()
        double_fit.do_fit()

        with self.subTest("parameter_values changed during fit"):
            for _parameter_value in regular_fit.parameter_values:
                assert _parameter_value != 1.0
            for _parameter_value in double_fit.parameter_values:
                assert _parameter_value != 1.0
        self._assert_fit_results_equal(
            regular_fit, double_fit, rtol=rtol, atol=atol,
            check_cost_function_value=check_cost_function_value, fit_2_is_double_fit=True)

    def _assert_properties_callable(self, fit):
        with self.subTest("data"):
            _ = fit.data
        #with self.subTest("data_error"):
        #    _ = fit.data_error
        #with self.subTest("data_cov_mat"):
        #    _ = fit.data_cov_mat
        #with self.subTest("data_cov_mat_inverse"):
        #    _ = fit.data_cov_mat_inverse
        #with self.subTest("data_cor_mat"):
        #    _ = fit.data_cor_mat
        with self.subTest("data_container"):
            _ = fit.data_container
        with self.subTest("model"):
            _ = fit.model
        #with self.subTest("model_error"):
        #    _ = fit.model_error
        #with self.subTest("model_cov_mat"):
        #    _ = fit.model_cov_mat
        #with self.subTest("model_cov_mat_inverse"):
        #    _ = fit.model_cov_mat_inverse
        #with self.subTest("model_cor_mat"):
        #    _ = fit.model_cor_mat
        with self.subTest("total_error"):
            _ = fit.total_error
        with self.subTest("total_cov_mat"):
            _ = fit.total_cov_mat
        with self.subTest("total_cov_mat_inverse"):
            _ = fit.total_cov_mat_inverse
        with self.subTest("total_cor_mat"):
            _ = fit.total_cor_mat
        with self.subTest("model_function"):
            _ = fit.model_function
        with self.subTest("model_label"):
            _ = fit.model_label
        with self.subTest("parameter_values"):
            _ = fit.parameter_values
        with self.subTest("parameter_values"):
            _ = fit.parameter_names
        with self.subTest("parameter_errors"):
            _ = fit.parameter_errors
        with self.subTest("asymmetric_parameter_errors"):
            _ = fit.asymmetric_parameter_errors
        with self.subTest("parameter_cov_mat"):
            _ = fit.parameter_cov_mat
        with self.subTest("parameter_cor_mat"):
            _ = fit.parameter_cor_mat
        with self.subTest("parameter_name_value_dict"):
            _ = fit.parameter_name_value_dict
        with self.subTest("parameter_constraints"):
            _ = fit.parameter_constraints
        with self.subTest("cost_function_value"):
            _ = fit.cost_function_value
        with self.subTest("data_size"):
            _ = fit.data_size
        with self.subTest("has_model_errors"):
            _ = fit.has_model_errors
        with self.subTest("has_data_errors"):
            _ = fit.has_data_errors
        with self.subTest("has_errors"):
            _ = fit.has_errors
        with self.subTest("model_count"):
            _ = fit.model_count
        with self.subTest("did_fit"):
            _ = fit.did_fit
        with self.subTest("ndf"):
            _ = fit.ndf
        with self.subTest("goodness_of_fit"):
            _ = fit.goodness_of_fit
        with self.subTest("dynamic_error_algorithm"):
            _ = fit.dynamic_error_algorithm
        with self.subTest("chi2_probability"):
            _ = fit.chi2_probability

    @staticmethod
    def _cost_function_custom_fit(a, b, c, xy_data, err_x=0.01, err_y=0.01):
        x_data, y_data = xy_data
        y_model = quadratic_model(x_data, a, b, c)
        residuals = y_data - y_model
        error_squared = err_y ** 2 + (quadratic_model_derivative(x_data, a, b, c) * err_x) ** 2
        chi2 = np.sum(residuals ** 2 / error_squared)
        det_cost = np.sum(np.log(error_squared))
        return chi2 + det_cost

    def _get_custom_fit(self, xy_data):
        _err_x = 0.01 if self._x_error else 0.0

        def cost_function(a, b, c):
            return TestMultiFit._cost_function_custom_fit(
                a, b, c, xy_data=xy_data, err_x=_err_x)

        return CustomFit(cost_function, minimizer=self._minimizer)

    def _set_custom_fits(self):
        self._xy_data = TestMultiFit._get_xy_data()
        self._fit_custom_all = self._get_custom_fit(self._xy_data)
        self._split_data_1, self._split_data_2 = TestMultiFit._split_data(self._xy_data, axis=1)
        self._fit_custom_all_multi = MultiFit(
            [self._get_custom_fit(self._xy_data)])
        self._fit_custom_split_multi = MultiFit(fit_list=[
            self._get_custom_fit(self._split_data_1),
            self._get_custom_fit(self._split_data_2)
        ])
        self._fit_custom_split_1 = self._get_custom_fit(self._split_data_1)
        self._fit_custom_split_1_multi = MultiFit(
            [self._get_custom_fit(self._split_data_1)])
        self._fit_custom_split_2 = self._get_custom_fit(self._split_data_2)
        self._fit_custom_split_2_multi = MultiFit(
            [self._get_custom_fit(self._split_data_2)])
        self._fit_custom_all_double = MultiFit([
            self._get_custom_fit(self._xy_data),
            self._get_custom_fit(self._xy_data)
        ])

    @staticmethod
    def _get_hist_data(loc=2.5, scale=0.5):
        return np.random.normal(loc=loc, scale=scale, size=1000)

    @staticmethod
    def _norm_cdf(x, mu, sigma):
        return norm.cdf(x=x, loc=mu, scale=sigma)

    def _get_hist_fit(self, fill_data):
        return HistFit(
            HistContainer(
                n_bins=100,
                bin_range=(1.5, 3.5),
                fill_data=fill_data
            ),
            bin_evaluation=TestMultiFit._norm_cdf,
            minimizer=self._minimizer
        )

    def _set_hist_fits(self):
        _raw_data = TestMultiFit._get_hist_data()
        self._fit_hist_all = self._get_hist_fit(_raw_data)
        _split_data_1, _split_data_2 = self._split_data(_raw_data)
        self._fit_hist_all_multi = MultiFit([self._get_hist_fit(_raw_data)])
        self._fit_hist_split_multi = MultiFit(fit_list=[
            self._get_hist_fit(_split_data_1),
            self._get_hist_fit(_split_data_2),
        ])
        self._fit_hist_split_1 = self._get_hist_fit(_split_data_1)
        self._fit_hist_split_1_multi = MultiFit(
            [self._get_hist_fit(_split_data_1)])
        self._fit_hist_split_2 = self._get_hist_fit(_split_data_2)
        self._fit_hist_split_2_multi = MultiFit(
            [self._get_hist_fit(_split_data_2)])

    @staticmethod
    def _get_indexed_data(err_y=0.01, a=1.2, b=2.3, c=3.4):
        _x_0 = np.linspace(start=-10, stop=10, num=101, endpoint=True)
        _y_0 = quadratic_model(_x_0, a, b, c)
        _y_jitter = np.random.normal(loc=0, scale=err_y, size=101)
        _y_data = _y_0 + _y_jitter

        return _y_data

    @staticmethod
    def quadratic_model_indexed_all(a, b, c):
        _x = np.linspace(start=-10, stop=10, num=101, endpoint=True)
        return quadratic_model(_x, a, b, c)

    @staticmethod
    def quadratic_model_indexed_split_1(a, b, c):
        _x = np.linspace(start=-10, stop=0, num=50, endpoint=False)
        return quadratic_model(_x, a, b, c)

    @staticmethod
    def quadratic_model_indexed_split_2(a, b, c):
        _x = np.linspace(start=0, stop=10, num=51, endpoint=True)
        return quadratic_model(_x, a, b, c)

    def _get_indexed_fit(self, data, model_function):
        _err = dict(err_val=0.1, relative=True, reference="model") \
            if self._relative_model_error else dict(err_val=0.1)
        _indexed_fit = IndexedFit(
            data=data,
            model_function=model_function,
            cost_function='chi2',
            minimizer=self._minimizer
        )
        _indexed_fit.add_error(**_err)
        return _indexed_fit

    def _set_indexed_fits(self):
        _data = TestMultiFit._get_indexed_data()
        self._fit_indexed_all = self._get_indexed_fit(
            _data, TestMultiFit.quadratic_model_indexed_all)
        _split_data_1 = _data[:50]
        _split_data_2 = _data[50:]
        self._fit_indexed_all_multi = MultiFit([self._get_indexed_fit(
            _data, TestMultiFit.quadratic_model_indexed_all)])
        self._fit_indexed_split_multi = MultiFit(fit_list=[
            self._get_indexed_fit(
                _split_data_1, TestMultiFit.quadratic_model_indexed_split_1),
            self._get_indexed_fit(
                _split_data_2, TestMultiFit.quadratic_model_indexed_split_2)
        ])
        self._fit_indexed_split_1 = self._get_indexed_fit(
            _split_data_1, TestMultiFit.quadratic_model_indexed_split_1)
        self._fit_indexed_split_1_multi = MultiFit([self._get_indexed_fit(
            _split_data_1, TestMultiFit.quadratic_model_indexed_split_1)])
        self._fit_indexed_split_2 = self._get_indexed_fit(
            _split_data_2, TestMultiFit.quadratic_model_indexed_split_2)
        self._fit_indexed_split_2_multi = MultiFit([self._get_indexed_fit(
            _split_data_2, TestMultiFit.quadratic_model_indexed_split_2)])

    @staticmethod
    def _get_xy_data(err_x=0.01, err_y=0.01, a=0.2, b=2.3, c=13.4, relative_model_error=False):
        _x_0 = np.linspace(start=-10, stop=10, num=25, endpoint=True)
        _x_jitter = np.random.normal(loc=0, scale=err_x, size=25)
        _x_data = _x_0 + _x_jitter

        _y_0 = quadratic_model(_x_data, a, b, c)
        _y_scale = err_y if not relative_model_error else err_y * _y_0
        _y_jitter = np.random.normal(loc=0, scale=_y_scale, size=25)
        _y_data = _y_0 + _y_jitter

        return np.array([_x_data, _y_data])

    def _get_xy_fit(self, xy_data):
        _err_x = dict(axis="x", err_val=0.01, name="x") if self._x_error is not None else None
        _err_y = dict(axis="y", err_val=0.01, relative=True, reference="model", name="y") \
            if self._relative_model_error else dict(axis="y", err_val=0.01, name="y")
        if self._x_error is None:
            _xy_fit = XYFit(
                xy_data=xy_data,
                model_function=quadratic_model,
                cost_function='chi2',
                minimizer=self._minimizer
            )
        else:
            _xy_fit = XYFit(
                xy_data=xy_data,
                model_function=quadratic_model,
                cost_function='chi2',
                minimizer=self._minimizer,
                dynamic_error_algorithm=self._x_error
            )
        if _err_x is not None:
            _xy_fit.add_error(**_err_x)
        _xy_fit.add_error(**_err_y)
        return _xy_fit

    def _set_xy_fits(self):
        self._xy_data = TestMultiFit._get_xy_data(relative_model_error=self._relative_model_error)
        self._fit_xy_all = self._get_xy_fit(self._xy_data)
        self._split_data_1, self._split_data_2 = TestMultiFit._split_data(self._xy_data, axis=1)
        self._fit_xy_all_multi = MultiFit(
            [self._get_xy_fit(self._xy_data)])
        self._fit_xy_split_multi = MultiFit(fit_list=[
            self._get_xy_fit(self._split_data_1),
            self._get_xy_fit(self._split_data_2)
        ])
        self._fit_xy_split_1 = self._get_xy_fit(self._split_data_1)
        self._fit_xy_split_1_multi = MultiFit(
            [self._get_xy_fit(self._split_data_1)])
        self._fit_xy_split_2 = self._get_xy_fit(self._split_data_2)
        self._fit_xy_split_2_multi = MultiFit(
            [self._get_xy_fit(self._split_data_2)])
        self._fit_xy_all_double = MultiFit([
            self._get_xy_fit(self._xy_data),
            self._get_xy_fit(self._xy_data)
        ])

    def _set_expected_xy_par_values(self):
        _x_err_val = 0.01 if self._x_error is not None else None
        # Note that when self._relative_model_error is True data errors are added here.
        self._expected_parameters_all = calculate_expected_fit_parameters_xy(
            x_data=self._xy_data[0], y_data=self._xy_data[1], model_function=quadratic_model,
            y_error=0.01, initial_parameter_values=[1.0, 1.0, 1.0], x_error=_x_err_val,
            model_function_derivative=quadratic_model_derivative,
            relative_model_y_error=self._relative_model_error
        )
        self._expected_parameters_split_1 = calculate_expected_fit_parameters_xy(
            x_data=self._split_data_1[0], y_data=self._split_data_1[1],
            model_function=quadratic_model,
            y_error=0.01, initial_parameter_values=[1.0, 1.0, 1.0], x_error=_x_err_val,
            model_function_derivative=quadratic_model_derivative,
            relative_model_y_error=self._relative_model_error
        )
        self._expected_parameters_split_2 = calculate_expected_fit_parameters_xy(
            x_data=self._split_data_2[0], y_data=self._split_data_2[1],
            model_function=quadratic_model,
            y_error=0.01, initial_parameter_values=[1.0, 1.0, 1.0], x_error=_x_err_val,
            model_function_derivative=quadratic_model_derivative,
            relative_model_y_error=self._relative_model_error
        )

    def setUp(self):
        np.random.seed(1337)
        self._minimizer = "scipy"
        self._relative_model_error = False
        self._x_error = None


class TestMultiFitIntegrityCustom(TestMultiFit):

    def setUp(self):
        TestMultiFit.setUp(self)

    def _assert_fits_match_expectation(self, atol, rtol):
        self._fit_custom_all.do_fit()
        self._assert_values_equal(
            "all", self._expected_parameters_all, self._fit_custom_all.parameter_values,
            atol=atol, rtol=rtol)
        self._fit_custom_all_multi.do_fit()
        self._assert_values_equal(
            "all", self._expected_parameters_all, self._fit_custom_all_multi.parameter_values,
            atol=atol, rtol=rtol)
        self._fit_custom_split_multi.do_fit()
        self._assert_values_equal(
            "all", self._expected_parameters_all, self._fit_custom_split_multi.parameter_values,
            atol=atol, rtol=rtol)
        self._fit_custom_split_1.do_fit()
        self._assert_values_equal(
            "all", self._expected_parameters_split_1, self._fit_custom_split_1.parameter_values,
            atol=atol, rtol=rtol)
        self._fit_custom_split_1_multi.do_fit()
        self._assert_values_equal(
            "all", self._expected_parameters_split_1,
            self._fit_custom_split_1_multi.parameter_values,
            atol=atol, rtol=rtol)
        self._fit_custom_split_2.do_fit()
        self._assert_values_equal(
            "all", self._expected_parameters_split_2, self._fit_custom_split_2.parameter_values,
            atol=atol, rtol=rtol)
        self._fit_custom_split_2_multi.do_fit()
        self._assert_values_equal(
            "all", self._expected_parameters_split_2,
            self._fit_custom_split_2_multi.parameter_values,
            atol=0, rtol=rtol)

    def test_properties_callable(self):
        self._set_custom_fits()
        self._assert_properties_callable(self._fit_custom_all_multi)

    def test_parameter_values_match_expectation(self):
        self._set_custom_fits()
        self._set_expected_xy_par_values()
        self._assert_fits_match_expectation(atol=0, rtol=1e-6)

    def test_split_fit_integrity_simple(self):
        self._set_custom_fits()
        self._assert_fits_valid_and_equal(self._fit_custom_all, self._fit_custom_all_multi)
        self._assert_fits_valid_and_equal(self._fit_custom_split_1, self._fit_custom_split_1_multi)
        self._assert_fits_valid_and_equal(self._fit_custom_split_2, self._fit_custom_split_2_multi)

    def test_split_fit_vs_regular_fit(self):
        self._set_custom_fits()
        self._assert_fits_valid_and_equal(self._fit_custom_all, self._fit_custom_split_multi)

    def test_double_fit_vs_regular_fit(self):
        self._set_custom_fits()
        self._assert_fits_valid_and_equal_double(self._fit_custom_all, self._fit_custom_all_double)

    def test_parameter_values_match_expectation_nonlinear_x_error(self):
        self._x_error = "nonlinear"
        self._set_custom_fits()
        self._set_expected_xy_par_values()
        self._assert_fits_match_expectation(atol=0, rtol=5e-4)


class TestMultiFitIntegrityHist(TestMultiFit):

    def setUp(self):
        TestMultiFit.setUp(self)

    def test_properties_callable(self):
        self._set_hist_fits()
        self._assert_properties_callable(self._fit_hist_all_multi)

    def test_split_fit_integrity_simple_scipy(self):
        self._set_hist_fits()
        self._assert_fits_valid_and_equal(self._fit_hist_all, self._fit_hist_all_multi)
        self._assert_fits_valid_and_equal(self._fit_hist_split_1, self._fit_hist_split_1_multi)
        self._assert_fits_valid_and_equal(self._fit_hist_split_2, self._fit_hist_split_2_multi)

    def test_split_fit_vs_regular_fit_scipy(self):
        self._set_hist_fits()
        self._assert_fits_valid_and_equal(
            self._fit_hist_all, self._fit_hist_split_multi, check_cost_function_value=False
        )


class TestMultiFitIntegrityIndexed(TestMultiFit):

    def setUp(self):
        TestMultiFit.setUp(self)

    def test_properties_callable(self):
        self._set_indexed_fits()
        self._assert_properties_callable(self._fit_indexed_all_multi)

    def test_split_fit_integrity_simple_scipy(self):
        self._set_indexed_fits()
        self._assert_fits_valid_and_equal(self._fit_indexed_all, self._fit_indexed_all_multi)
        self._assert_fits_valid_and_equal(
            self._fit_indexed_split_1, self._fit_indexed_split_1_multi)
        self._assert_fits_valid_and_equal(
            self._fit_indexed_split_2, self._fit_indexed_split_2_multi)

    def test_split_fit_vs_regular_fit_scipy(self):
        self._set_indexed_fits()
        self._assert_fits_valid_and_equal(self._fit_indexed_all, self._fit_indexed_split_multi)

    def test_split_fit_integrity_simple_scipy_with_relative_error(self):
        self._relative_model_error = True
        self._set_indexed_fits()
        self._assert_fits_valid_and_equal(
            self._fit_indexed_all, self._fit_indexed_all_multi)
        self._assert_fits_valid_and_equal(
            self._fit_indexed_split_1, self._fit_indexed_split_1_multi)
        self._assert_fits_valid_and_equal(
            self._fit_indexed_split_2, self._fit_indexed_split_2_multi)

    def test_split_fit_vs_regular_fit_scipy_with_relative_error(self):
        self._relative_model_error = True
        self._set_indexed_fits()
        self._assert_fits_valid_and_equal(
            self._fit_indexed_all, self._fit_indexed_split_multi)


class TestMultiFitIntegrityXY(TestMultiFit):

    def setUp(self):
        TestMultiFit.setUp(self)

    def _assert_fits_match_expectation(self, atol, rtol):
        self._fit_xy_all.do_fit()
        self._assert_values_equal(
            "all", self._fit_xy_all.parameter_values, self._expected_parameters_all,
            atol=atol, rtol=rtol)
        self._fit_xy_all_multi.do_fit()
        self._assert_values_equal(
            "all", self._fit_xy_all_multi.parameter_values, self._expected_parameters_all,
            atol=atol, rtol=rtol)
        self._fit_xy_split_multi.do_fit()
        self._assert_values_equal(
            "all", self._fit_xy_split_multi.parameter_values, self._expected_parameters_all,
            atol=atol, rtol=rtol)
        self._fit_xy_split_1.do_fit()
        self._assert_values_equal(
            "all", self._fit_xy_split_1.parameter_values, self._expected_parameters_split_1,
            atol=atol, rtol=rtol)
        self._fit_xy_split_1_multi.do_fit()
        self._assert_values_equal(
            "all", self._fit_xy_split_1_multi.parameter_values, self._expected_parameters_split_1,
            atol=atol, rtol=rtol)
        self._fit_xy_split_2.do_fit()
        self._assert_values_equal(
            "all", self._fit_xy_split_2.parameter_values, self._expected_parameters_split_2,
            atol=atol, rtol=rtol)
        self._fit_xy_split_2_multi.do_fit()
        self._assert_values_equal(
            "all", self._fit_xy_split_2_multi.parameter_values, self._expected_parameters_split_2,
            atol=0, rtol=rtol)

    def test_properties_callable(self):
        self._set_xy_fits()
        self._assert_properties_callable(self._fit_xy_all_multi)

    def test_parameter_values_match_expectation(self):
        self._set_xy_fits()
        self._set_expected_xy_par_values()
        self._assert_fits_match_expectation(atol=0, rtol=1e-6)

    def test_parameter_values_match_expectation_relative_y_error_nonlinear(self):
        self._x_error = "nonlinear"
        self._relative_model_error = True
        self._set_xy_fits()
        self._set_expected_xy_par_values()
        self._assert_fits_match_expectation(atol=0, rtol=1e-4)

    def test_parameter_values_match_expectation_relative_y_error_iterative(self):
        self._x_error = "iterative"
        self._relative_model_error = True
        self._set_xy_fits()
        self._set_expected_xy_par_values()
        self._assert_fits_match_expectation(atol=0, rtol=1e-4)

    def test_parameter_values_match_expectation_nonlinear_x_error(self):
        self._x_error = "nonlinear"
        self._set_xy_fits()
        self._set_expected_xy_par_values()
        self._assert_fits_match_expectation(atol=0, rtol=5e-4)

    def test_parameter_values_match_expectation_iterative_linear_x_error(self):
        self._x_error = "iterative"
        self._set_xy_fits()
        self._set_expected_xy_par_values()
        self._assert_fits_match_expectation(atol=0, rtol=5e-4)

    def test_split_fit_integrity_simple(self):
        self._set_xy_fits()
        self._assert_fits_valid_and_equal(self._fit_xy_all, self._fit_xy_all_multi)
        self._assert_fits_valid_and_equal(self._fit_xy_split_1, self._fit_xy_split_1_multi)
        self._assert_fits_valid_and_equal(self._fit_xy_split_2, self._fit_xy_split_2_multi)

    def test_split_fit_vs_regular_fit(self):
        self._set_xy_fits()
        self._assert_fits_valid_and_equal(self._fit_xy_all, self._fit_xy_split_multi)

    def test_split_fit_vs_regular_fit_with_nonlinear_relative_model_error(self):
        self._x_error = "nonlinear"
        self._relative_model_error = True
        self._set_xy_fits()
        self._assert_fits_valid_and_equal(self._fit_xy_all, self._fit_xy_split_multi)

    def test_split_fit_vs_regular_fit_with_iterative_relative_model_error(self):
        self._x_error = "iterative"
        self._relative_model_error = True
        self._set_xy_fits()
        self._assert_fits_valid_and_equal(self._fit_xy_all, self._fit_xy_split_multi)

    def test_double_fit_vs_regular_fit(self):
        self._set_xy_fits()
        self._assert_fits_valid_and_equal_double(self._fit_xy_all, self._fit_xy_all_double)

    def test_split_fit_vs_regular_fit_with_nonlinear_x_error(self):
        self._x_error = "nonlinear"
        self._set_xy_fits()
        self._assert_fits_valid_and_equal(self._fit_xy_all, self._fit_xy_split_multi)

    def test_split_fit_vs_regular_fit_with_iterative_x_error(self):
        self._x_error = "iterative"
        self._set_xy_fits()
        self._assert_fits_valid_and_equal(self._fit_xy_all, self._fit_xy_split_multi)
